/****************************************************************************
**
** Copyright (C) 2019 - 2022 Trevor SANDY. All rights reserved.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  Please review the following information to ensure GNU
** General Public Licensing requirements will be met:
** http://www.trolltech.com/products/qt/opensource.html
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

/****************************************************************************
 *
 * This class encapsulates the native renderer routines.
 *
 * Please see lpub.h for an overall description of how the files in LPub
 * make up the LPub program.
 *
 ***************************************************************************/

#include "lpub.h"
#include "editwindow.h"
#include "lpub_preferences.h"
#include "messageboxresizable.h"
#include "step.h"
#include "metagui.h"
#include "paths.h"
#include "lpub_qtcompat.h"

#include "lc_viewwidget.h"
#include "lc_previewwidget.h"
#include "pieceinf.h"
#include "lc_library.h"
#include "project.h"
#include "lc_mainwindow.h"

#include "lc_qutils.h"
#include "lc_profile.h"
#include "light.h"
#include "camera.h"
#include "piece.h"
#include "group.h"
#include "lc_view.h"
#include "application.h"
#include "lc_partselectionwidget.h"
#include "lc_setsdatabasedialog.h"

const QStringList BuildModChangeTriggers = QStringList()
        << "Deleting" << "Adding Piece"
        << "Move"     << "Rotate"
        << "Moving"   << "Rotating"
        << "Painting" << "New Model"
        << "Cutting"  << "Pasting"
        << "Removing Step" << "Inserting Step"
        << "Duplicating Pieces" << "New Model"
        << "Ungrouping" << "Grouping" << "Inlining"
           ;

void Gui::create3DActions()
{
    gMainWindow->mActions[LC_VIEW_PREFERENCES]->setObjectName("PreferencesAct.4");
    lpub->actions.insert("PreferencesAct.4", Action(QStringLiteral("3DViewer.Preferences"), gMainWindow->mActions[LC_VIEW_PREFERENCES]));

    gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]->setObjectName("PointLightAct.4");
    lpub->actions.insert("PointLightAct.4", Action(QStringLiteral("3DViewer.Tools.Lights.PointLight"), gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]));

    gMainWindow->mActions[LC_EDIT_ACTION_SUNLIGHT]->setObjectName("SunLightAct.4");
    lpub->actions.insert("SunLightAct.4", Action(QStringLiteral("3DViewer.Tools.Lights.SunLight"), gMainWindow->mActions[LC_EDIT_ACTION_SUNLIGHT]));

    gMainWindow->mActions[LC_EDIT_ACTION_SPOTLIGHT]->setObjectName("SpotLightAct.4");
    lpub->actions.insert("SpotLightAct.4", Action(QStringLiteral("3DViewer.Tools.Lights.SpotLight"), gMainWindow->mActions[LC_EDIT_ACTION_SPOTLIGHT]));

    gMainWindow->mActions[LC_EDIT_ACTION_AREALIGHT]->setObjectName("AreaLightAct.4");
    lpub->actions.insert("AreaLightAct.4", Action(QStringLiteral("3DViewer.Tools.Lights.AreaLight"), gMainWindow->mActions[LC_EDIT_ACTION_AREALIGHT]));

    gMainWindow->mActions[LC_FILE_SAVE_IMAGE]->setObjectName("SaveImageAct.4");
    lpub->actions.insert("SaveImageAct.4", Action(QStringLiteral("3DViewer.Save Image"), gMainWindow->mActions[LC_FILE_SAVE_IMAGE]));

    gMainWindow->mActions[LC_FILE_EXPORT_3DS]->setObjectName("Export3dsAct.4");
    lpub->actions.insert("Export3dsAct.4", Action(QStringLiteral("3DViewer.Export As.3DS"), gMainWindow->mActions[LC_FILE_EXPORT_3DS]));

    gMainWindow->mActions[LC_FILE_EXPORT_COLLADA]->setObjectName("ExportColladaAct.4");
    lpub->actions.insert("ExportColladaAct.4", Action(QStringLiteral("3DViewer.Export As.COLLADA"), gMainWindow->mActions[LC_FILE_EXPORT_COLLADA]));

    gMainWindow->mActions[LC_FILE_EXPORT_WAVEFRONT]->setObjectName("ExportWavefrontAct.4");
    lpub->actions.insert("ExportWavefrontAct.4", Action(QStringLiteral("3DViewer.Export As.Wavefront"), gMainWindow->mActions[LC_FILE_EXPORT_WAVEFRONT]));

    gMainWindow->mActions[LC_EDIT_SNAP_MOVE_TOGGLE]->setObjectName("EditSnapMoveToggleAct.4");
    lpub->actions.insert("EditSnapMoveToggleAct.4", Action(QStringLiteral("3DViewer.Tools.Snap.Move Snap Enabled"), gMainWindow->mActions[LC_EDIT_SNAP_MOVE_TOGGLE]));

    gMainWindow->mActions[LC_EDIT_SNAP_ANGLE_TOGGLE]->setObjectName("EditSnapAngleToggleAct.4");
    lpub->actions.insert("EditSnapAngleToggleAct.4", Action(QStringLiteral("3DViewer.Tools.SnapAngle.Rotation Snap Enabled"), gMainWindow->mActions[LC_EDIT_SNAP_ANGLE_TOGGLE]));

    gMainWindow->mActions[LC_EDIT_CUT]->setObjectName("EditCutAct.4");
    lpub->actions.insert("EditCutAct.4", Action(QStringLiteral("3DViewer.Cut"), gMainWindow->mActions[LC_EDIT_CUT]));

    gMainWindow->mActions[LC_EDIT_COPY]->setObjectName("EditCopyAct.4");
    lpub->actions.insert("EditCopyAct.4", Action(QStringLiteral("3DViewer.Copy"), gMainWindow->mActions[LC_EDIT_COPY]));

    gMainWindow->mActions[LC_EDIT_PASTE]->setObjectName("EditPasteAct.4");
    lpub->actions.insert("EditPasteAct.4", Action(QStringLiteral("3DViewer.Paste"), gMainWindow->mActions[LC_EDIT_PASTE]));

    gMainWindow->mActions[LC_PIECE_DELETE]->setObjectName("PiectDeleteAct.4");
    lpub->actions.insert("PiectDeleteAct.4", Action(QStringLiteral("3DViewer.Delete"), gMainWindow->mActions[LC_PIECE_DELETE]));

    gMainWindow->mActions[LC_PIECE_PAINT_SELECTED]->setObjectName("PaintSelectedAct.4");
    lpub->actions.insert("PaintSelectedAct.4", Action(QStringLiteral("3DViewer.Paint Selected"), gMainWindow->mActions[LC_PIECE_PAINT_SELECTED]));

    gMainWindow->mActions[LC_PIECE_EDIT_SELECTED_SUBMODEL]->setObjectName("EditSelectedSubmodelAct.4");
    lpub->actions.insert("EditSelectedSubmodelAct.4", Action(QStringLiteral("3DViewer.Edit Selected Submodel"), gMainWindow->mActions[LC_PIECE_EDIT_SELECTED_SUBMODEL]));

    gMainWindow->mActions[LC_PIECE_EDIT_END_SUBMODEL]->setObjectName("EndSubmodelEditAct.4");
    lpub->actions.insert("EndSubmodelEditAct.4", Action(QStringLiteral("3DViewer.End Submodel Editing"), gMainWindow->mActions[LC_PIECE_EDIT_END_SUBMODEL]));

    gMainWindow->mActions[LC_EDIT_ACTION_CLEAR_TRANSFORM]->setObjectName("EditClearTransformAct.4");
    lpub->actions.insert("EditClearTransformAct.4", Action(QStringLiteral("3DViewer.Tools.Clear Transform"), gMainWindow->mActions[LC_EDIT_ACTION_CLEAR_TRANSFORM]));

    gMainWindow->mActions[LC_PIECE_HIDE_SELECTED]->setObjectName("HideSelectedAct.4");
    lpub->actions.insert("HideSelectedAct.4", Action(QStringLiteral("3DViewer.Hide Selected"), gMainWindow->mActions[LC_PIECE_HIDE_SELECTED]));

    gMainWindow->mActions[LC_PIECE_UNHIDE_SELECTED]->setObjectName("UnhideSelectedAct.4");
    lpub->actions.insert("UnhideSelectedAct.4", Action(QStringLiteral("3DViewer.Unhide Selected"), gMainWindow->mActions[LC_PIECE_UNHIDE_SELECTED]));

    gMainWindow->mActions[LC_PIECE_HIDE_UNSELECTED]->setObjectName("HideUnselectedAct.4");
    lpub->actions.insert("HideUnselectedAct.4", Action(QStringLiteral("3DViewer.Hide Unselected"), gMainWindow->mActions[LC_PIECE_HIDE_UNSELECTED]));

    gMainWindow->mActions[LC_PIECE_UNHIDE_ALL]->setObjectName("UnhideAllAct.4");
    lpub->actions.insert("UnhideAllAct.4", Action(QStringLiteral("3DViewer.Unhide All"), gMainWindow->mActions[LC_PIECE_UNHIDE_ALL]));

    gMainWindow->mActions[LC_EDIT_ACTION_SELECT]->setObjectName("EditSelectAct.4");
    lpub->actions.insert("EditSelectAct.4", Action(QStringLiteral("3DViewer.Tools.Select"), gMainWindow->mActions[LC_EDIT_ACTION_SELECT]));

    gMainWindow->mActions[LC_EDIT_ACTION_ROTATE]->setObjectName("EditRotateAct.4");
    lpub->actions.insert("EditRotateAct.4", Action(QStringLiteral("3DViewer.Tools.Rotate"), gMainWindow->mActions[LC_EDIT_ACTION_ROTATE]));

    gMainWindow->mActions[LC_EDIT_ACTION_MOVE]->setObjectName("EditMoveAct.4");
    lpub->actions.insert("EditMoveAct.4", Action(QStringLiteral("3DViewer.Tools.Move"), gMainWindow->mActions[LC_EDIT_ACTION_MOVE]));

    gMainWindow->mActions[LC_EDIT_ACTION_DELETE]->setObjectName("EditDeleteAct.4");
    lpub->actions.insert("EditDeleteAct.4", Action(QStringLiteral("3DViewer.Tools.Delete"), gMainWindow->mActions[LC_EDIT_ACTION_DELETE]));

    gMainWindow->mActions[LC_EDIT_ACTION_PAINT]->setObjectName("EditPaintAct.4");
    lpub->actions.insert("EditPaintAct.4", Action(QStringLiteral("3DViewer.Tools.Paint"), gMainWindow->mActions[LC_EDIT_ACTION_PAINT]));

    gMainWindow->mActions[LC_EDIT_ACTION_COLOR_PICKER]->setObjectName("EditColorPickerAct.4");
    lpub->actions.insert("EditColorPickerAct.4", Action(QStringLiteral("3DViewer.Tools.Color Picker"), gMainWindow->mActions[LC_EDIT_ACTION_COLOR_PICKER]));

    gMainWindow->mActions[LC_EDIT_ACTION_ROTATESTEP]->setObjectName("EditRotateStepAct.4");
    lpub->actions.insert("EditRotateStepAct.4", Action(QStringLiteral("3DViewer.Tools.Apply ROTSTEP"), gMainWindow->mActions[LC_EDIT_ACTION_ROTATESTEP]));

    gMainWindow->mActions[LC_EDIT_ACTION_INSERT]->setObjectName("EditInsertAct.4");
    lpub->actions.insert("EditInsertAct.4", Action(QStringLiteral("3DViewer.Tools.Insert"), gMainWindow->mActions[LC_EDIT_ACTION_INSERT]));

    gMainWindow->mActions[LC_EDIT_ACTION_CAMERA]->setObjectName("EditCameraAct.4");
    lpub->actions.insert("EditCameraAct.4", Action(QStringLiteral("3DViewer.Tools.Camera"), gMainWindow->mActions[LC_EDIT_ACTION_CAMERA]));

    gMainWindow->mActions[LC_VIEW_LOOK_AT]->setObjectName("ViewLookAtAct.4");
    lpub->actions.insert("ViewLookAtAct.4", Action(QStringLiteral("3DViewer.Tools.Look At"), gMainWindow->mActions[LC_VIEW_LOOK_AT]));

    gMainWindow->mActions[LC_VIEW_ZOOM_EXTENTS]->setObjectName("ViewZoomExtentsAct.4");
    lpub->actions.insert("ViewZoomExtentsAct.4", Action(QStringLiteral("3DViewer.Tools.Zoom Extents"), gMainWindow->mActions[LC_VIEW_ZOOM_EXTENTS]));

    gMainWindow->mActions[LC_EDIT_ACTION_ZOOM]->setObjectName("EditZoomAct.4");
    lpub->actions.insert("EditZoomAct.4", Action(QStringLiteral("3DViewer.Tools.Zoom"), gMainWindow->mActions[LC_EDIT_ACTION_ZOOM]));

    gMainWindow->mActions[LC_EDIT_ACTION_ZOOM_REGION]->setObjectName("EditZoomRegionAct.4");
    lpub->actions.insert("EditZoomRegionAct.4", Action(QStringLiteral("3DViewer.Tools.Zoom Region"), gMainWindow->mActions[LC_EDIT_ACTION_ZOOM_REGION]));

    gMainWindow->mActions[LC_EDIT_ACTION_ROTATE_VIEW]->setObjectName("EditRotateViewAct.4");
    lpub->actions.insert("EditRotateViewAct.4", Action(QStringLiteral("3DViewer.Tools.Rotate View"), gMainWindow->mActions[LC_EDIT_ACTION_ROTATE_VIEW]));

    gMainWindow->mActions[LC_EDIT_ACTION_PAN]->setObjectName("EditPanAct.4");
    lpub->actions.insert("EditPanAct.4", Action(QStringLiteral("3DViewer.Tools.Pan"), gMainWindow->mActions[LC_EDIT_ACTION_PAN]));

    gMainWindow->mActions[LC_VIEW_SPLIT_HORIZONTAL]->setObjectName("ViewSplitHorizontalAct.4");
    lpub->actions.insert("ViewSplitHorizontalAct.4", Action(QStringLiteral("3DViewer.Split Horizontal"), gMainWindow->mActions[LC_VIEW_SPLIT_HORIZONTAL]));

    gMainWindow->mActions[LC_VIEW_SPLIT_VERTICAL]->setObjectName("ViewSplitVerticalAct.4");
    lpub->actions.insert("ViewSplitVerticalAct.4", Action(QStringLiteral("3DViewer.Split Vertical"), gMainWindow->mActions[LC_VIEW_SPLIT_VERTICAL]));

    gMainWindow->mActions[LC_VIEW_REMOVE_VIEW]->setObjectName("ViewRemoveViewAct.4");
    lpub->actions.insert("ViewRemoveViewAct.4", Action(QStringLiteral("3DViewer.Remove View"), gMainWindow->mActions[LC_VIEW_REMOVE_VIEW]));

    gMainWindow->mActions[LC_VIEW_RESET_VIEWS]->setObjectName("ViewResetViewsAct.4");
    lpub->actions.insert("ViewResetViewsAct.4", Action(QStringLiteral("3DViewer.Reset Views"), gMainWindow->mActions[LC_VIEW_RESET_VIEWS]));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]->setObjectName("ViewViewpointsHomeAct.4");
    lpub->actions.insert("ViewViewpointsHomeAct.4", Action(QStringLiteral("3DViewer.Tools.Viewpoints.Home"), gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_FRONT]->setObjectName("ViewViewpointsFrontAct.4");
    lpub->actions.insert("ViewViewpointsFrontAct.4", Action(QStringLiteral("3DViewer.Tools.Viewpoints.Front"), gMainWindow->mActions[LC_VIEW_VIEWPOINT_FRONT]));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BACK]->setObjectName("ViewViewpointsBackAct.4");
    lpub->actions.insert("ViewViewpointsBackAct.4", Action(QStringLiteral("3DViewer.Tools.Viewpoints.Back"), gMainWindow->mActions[LC_VIEW_VIEWPOINT_BACK]));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_LEFT]->setObjectName("ViewViewpointsLeftAct.4");
    lpub->actions.insert("ViewViewpointsLeftAct.4", Action(QStringLiteral("3DViewer.Tools.Viewpoints.Left"), gMainWindow->mActions[LC_VIEW_VIEWPOINT_LEFT]));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_RIGHT]->setObjectName("ViewViewpointsRightAct.4");
    lpub->actions.insert("ViewViewpointsRightAct.4", Action(QStringLiteral("3DViewer.Tools.Viewpoints.Right"), gMainWindow->mActions[LC_VIEW_VIEWPOINT_RIGHT]));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_TOP]->setObjectName("ViewViewpointsTopAct.4");
    lpub->actions.insert("ViewViewpointsTopAct.4", Action(QStringLiteral("3DViewer.Tools.Viewpoints.Top"), gMainWindow->mActions[LC_VIEW_VIEWPOINT_TOP]));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BOTTOM]->setObjectName("ViewViewpointsBottomAct.4");
    lpub->actions.insert("ViewViewpointsBottomAct.4", Action(QStringLiteral("3DViewer.Tools.Viewpoints.Bottom"), gMainWindow->mActions[LC_VIEW_VIEWPOINT_BOTTOM]));

    gMainWindow->mActions[LC_HELP_ABOUT]->setObjectName("HelpAboutAct.4");
    lpub->actions.insert("HelpAboutAct.4", Action(QStringLiteral("3DViewer.About"), gMainWindow->mActions[LC_HELP_ABOUT]));

    QIcon CreateBuildModIcon;
    CreateBuildModIcon.addFile(":/resources/buildmodcreate.png");
    CreateBuildModIcon.addFile(":/resources/buildmodcreate16.png");
    CreateBuildModAct = new QAction(CreateBuildModIcon,tr("Create Build Modification"),this);
    CreateBuildModAct->setObjectName("CreateBuildModAct.4");
    CreateBuildModAct->setStatusTip(tr("Create a new build modification for this step"));
    CreateBuildModAct->setShortcut(QStringLiteral("Shift+J"));
    lpub->actions.insert(CreateBuildModAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Create Build Modification"), CreateBuildModAct));
    connect(CreateBuildModAct, SIGNAL(triggered()), this, SLOT(createBuildModification()));

    QIcon UpdateBuildModIcon;
    UpdateBuildModIcon.addFile(":/resources/buildmodupdate.png");
    UpdateBuildModIcon.addFile(":/resources/buildmodupdate16.png");
    UpdateBuildModAct = new QAction(UpdateBuildModIcon,tr("Update Build Modification"),this);
    UpdateBuildModAct->setObjectName("UpdateBuildModAct.4");
    UpdateBuildModAct->setEnabled(false);
    UpdateBuildModAct->setStatusTip(tr("Commit changes to the current build modification"));
    UpdateBuildModAct->setShortcut(QStringLiteral("Shift+K"));
    lpub->actions.insert(UpdateBuildModAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Build Modification.Update Build Modification"), UpdateBuildModAct));
    connect(UpdateBuildModAct, SIGNAL(triggered()), this, SLOT(updateBuildModification()));

    QIcon ApplyBuildModIcon;
    ApplyBuildModIcon.addFile(":/resources/buildmodapply.png");
    ApplyBuildModIcon.addFile(":/resources/buildmodapply16.png");
    ApplyBuildModAct = new QAction(ApplyBuildModIcon,tr("Apply Build Modification..."),this);
    ApplyBuildModAct->setObjectName("ApplyBuildModAct.4");
    ApplyBuildModAct->setEnabled(false);
    ApplyBuildModAct->setStatusTip(tr("Apply existing build modification to this step"));
    lpub->actions.insert(ApplyBuildModAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Build Modification.Apply Build Modification"), ApplyBuildModAct));
    connect(ApplyBuildModAct, SIGNAL(triggered()), this, SLOT(applyBuildModification()));

    QIcon RemoveBuildModIcon;
    RemoveBuildModIcon.addFile(":/resources/buildmodremove.png");
    RemoveBuildModIcon.addFile(":/resources/buildmodremove16.png");
    RemoveBuildModAct = new QAction(RemoveBuildModIcon,tr("Remove Build Modification..."),this);
    RemoveBuildModAct->setObjectName("RemoveBuildModAct.4");
    RemoveBuildModAct->setEnabled(false);
    RemoveBuildModAct->setStatusTip(tr("Remove build modification from this step"));
    lpub->actions.insert(RemoveBuildModAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Build Modification.Remove Build Modification"), RemoveBuildModAct));
    connect(RemoveBuildModAct, SIGNAL(triggered()), this, SLOT(removeBuildModification()));

    QIcon LoadBuildModIcon;
    LoadBuildModIcon.addFile(":/resources/buildmodload.png");
    LoadBuildModIcon.addFile(":/resources/buildmodload16.png");
    LoadBuildModAct = new QAction(LoadBuildModIcon,tr("Load Build Modification..."),this);
    LoadBuildModAct->setObjectName("LoadBuildModAct.4");
    LoadBuildModAct->setEnabled(false);
    LoadBuildModAct->setStatusTip(tr("Load the step containing the selected build modification"));

    lpub->actions.insert(LoadBuildModAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Build Modification.Load Build Modification"), LoadBuildModAct));
    connect(LoadBuildModAct, SIGNAL(triggered()), this, SLOT(loadBuildModification()));

    QIcon DeleteBuildModIcon;
    DeleteBuildModIcon.addFile(":/resources/buildmoddelete.png");
    DeleteBuildModIcon.addFile(":/resources/buildmoddelete16.png");
    DeleteBuildModAct = new QAction(DeleteBuildModIcon,tr("Delete Build Modification..."),this);
    DeleteBuildModAct->setObjectName("DeleteBuildModAct.4");
    DeleteBuildModAct->setEnabled(false);
    DeleteBuildModAct->setStatusTip(tr("Delete selected build modification meta commands"));
    lpub->actions.insert(DeleteBuildModAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Build Modification.Text"), DeleteBuildModAct));
    connect(DeleteBuildModAct, SIGNAL(triggered()), this, SLOT(deleteBuildModification()));

    EnableBuildModAct = new QAction(tr("Build Modifications Enabled"),this);
    EnableBuildModAct->setObjectName("EnableBuildModAct.4");
    EnableBuildModAct->setStatusTip(tr("Enable build modification configuration on rotate action"));
    EnableBuildModAct->setCheckable(true);
    EnableBuildModAct->setChecked(lcGetProfileInt(LC_PROFILE_BUILD_MODIFICATION));
    lpub->actions.insert(EnableBuildModAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Rotate.Build Modifications Enabled"), EnableBuildModAct));
    connect(EnableBuildModAct, SIGNAL(triggered()), this, SLOT(enableBuildModification()));

    EnableRotstepRotateAct = new QAction(tr("ROTSTEP Rotate Enabled"),this);
    EnableRotstepRotateAct->setObjectName("EnableRotstepRotateAct.4");
    EnableRotstepRotateAct->setStatusTip(tr("Enable ROTSTEP configuration on rotate action"));
    EnableRotstepRotateAct->setCheckable(true);
    EnableRotstepRotateAct->setChecked(!lcGetProfileInt(LC_PROFILE_BUILD_MODIFICATION));
    lpub->actions.insert(EnableRotstepRotateAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Rotate.ROTSTEP Rotate Enabled"), EnableRotstepRotateAct));
    connect(EnableRotstepRotateAct, SIGNAL(triggered()), this, SLOT(enableBuildModification()));

    QIcon ApplyLightIcon;
    ApplyLightIcon.addFile(":/resources/applylightsettings.png");
    ApplyLightIcon.addFile(":/resources/applylightsettings_16.png");
    ApplyLightAct = new QAction(ApplyLightIcon,tr("Save Light Settings"),this);
    ApplyLightAct->setObjectName("ApplyLightAct.4");
    ApplyLightAct->setEnabled(false);
    ApplyLightAct->setStatusTip(tr("Save light settings to current step"));
    ApplyLightAct->setShortcut(QStringLiteral("Shift+I"));
    lpub->actions.insert(ApplyLightAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Lights.Save Light Settings"), ApplyLightAct));
    connect(ApplyLightAct, SIGNAL(triggered()), this, SLOT(applyLightSettings()));

    QIcon ApplyCameraIcon;
    ApplyCameraIcon.addFile(":/resources/applycamerasettings.png");
    ApplyCameraIcon.addFile(":/resources/applycamerasettings_16.png");
    ApplyCameraAct = new QAction(ApplyCameraIcon,tr("Save Camera Settings"),this);
    ApplyCameraAct->setObjectName("ApplyCameraAct.4");
    ApplyCameraAct->setStatusTip(tr("Save current camera settings to current step and regenerate image"));
    ApplyCameraAct->setShortcut(QStringLiteral("Shift+A"));
    lpub->actions.insert(ApplyCameraAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Camera.Save Camera Settings"), ApplyCameraAct));
    connect(ApplyCameraAct, SIGNAL(triggered()), this, SLOT(applyCameraSettings()));

    UseImageSizeAct = new QAction(tr("Use Image Size"),this);
    UseImageSizeAct->setObjectName("UseImageSizeAct.4");
    UseImageSizeAct->setStatusTip(tr("Use image width and height - you can also edit width and height in camera Properties tab"));
    UseImageSizeAct->setCheckable(true);
    UseImageSizeAct->setChecked(lcGetProfileInt(LC_PROFILE_USE_IMAGE_SIZE));
    lpub->actions.insert(UseImageSizeAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Camera.Use Image Size"), UseImageSizeAct));
    connect(UseImageSizeAct, SIGNAL(triggered()), this, SLOT(useImageSize()));

    AutoCenterSelectionAct = new QAction(tr("Look At Selection"),this);
    AutoCenterSelectionAct->setObjectName("AutoCenterSelectionAct.4");
    AutoCenterSelectionAct->setStatusTip(tr("Automatically rotate view so selected pieces are at center"));
    AutoCenterSelectionAct->setCheckable(true);
    AutoCenterSelectionAct->setChecked(lcGetProfileInt(LC_PROFILE_AUTO_CENTER_SELECTION));
    lpub->actions.insert(AutoCenterSelectionAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Camera.Look At Selection"), AutoCenterSelectionAct));
    connect(AutoCenterSelectionAct, SIGNAL(triggered()), this, SLOT(autoCenterSelection()));

    DefaultCameraPropertiesAct = new QAction(tr("Display Properties"),this);
    DefaultCameraPropertiesAct->setObjectName("DefaultCameraPropertiesAct.4");
    DefaultCameraPropertiesAct->setStatusTip(tr("Display default camera properties in Properties tab"));
    DefaultCameraPropertiesAct->setCheckable(true);
    DefaultCameraPropertiesAct->setChecked(GetPreferences().mDefaultCameraProperties);
    lpub->actions.insert(DefaultCameraPropertiesAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Camera.Display Properties"), DefaultCameraPropertiesAct));
    connect(DefaultCameraPropertiesAct, SIGNAL(triggered()), this, SLOT(showDefaultCameraProperties()));

    TransformAct = new QAction(tr("Transform"), this);
    TransformAct->setObjectName("TransformAct.4");
    TransformAct->setStatusTip(tr("Transform Options"));
    lpub->actions.insert(TransformAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Transform"), TransformAct));
    TransformAct->setIcon(QIcon(":/resources/edit_transform_relative.png"));

    MoveAct = new QAction(tr("Movement Snap"), this);
    MoveAct->setObjectName("MoveAct.4");
    MoveAct->setStatusTip(tr("Snap translations to fixed intervals"));
    MoveAct->setIcon(QIcon(":/resources/edit_snap_move.png"));
    lpub->actions.insert(MoveAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Movement Snap"), MoveAct));

    AngleAct = new QAction(tr("Rotation Snap"), this);
    AngleAct->setObjectName("AngleAct.4");
    AngleAct->setStatusTip(tr("Snap rotations to fixed intervals"));
    AngleAct->setIcon(QIcon(":/resources/edit_snap_angle.png"));
    lpub->actions.insert(AngleAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Rotation Snap"), AngleAct));

    QIcon ViewViewPointHomeIcon;
    ViewViewPointHomeIcon.addFile(":/resources/veiw_viewpoint_home.png");
    ViewViewPointHomeIcon.addFile(":/resources/veiw_viewpoint_home_16.png");
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]->setIcon(ViewViewPointHomeIcon);

    QIcon EditActionRotstepIcon;
    EditActionRotstepIcon.addFile(":/resources/edit_rotatestep.png");
    EditActionRotstepIcon.addFile(":/resources/edit_rotatestep_16.png");
    gMainWindow->mActions[LC_EDIT_ACTION_ROTATESTEP]->setIcon(EditActionRotstepIcon);

    gMainWindow->mActions[LC_EDIT_ACTION_CLEAR_TRANSFORM]->setIcon(QIcon(":/resources/clear_transform.png"));

    // Light icons
    LightGroupAct = new QAction(tr("Lights"), this);
    LightGroupAct->setObjectName("LightGroupAct.4");
    LightGroupAct->setToolTip(tr("Lights - Pointlight"));
    LightGroupAct->setIcon(gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]->icon());
    LightGroupAct->setStatusTip(gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]->statusTip());
    LightGroupAct->setProperty("CommandId", QVariant(LC_EDIT_ACTION_LIGHT));
    lpub->actions.insert(LightGroupAct->objectName(), Action(QStringLiteral("3DViewer.Tools.Lights.Lights"), LightGroupAct));
    // Light icons
    LightGroupAct = new QAction(tr("Lights"), this);
    LightGroupAct->setToolTip(tr("Lights - Pointlight"));
    LightGroupAct->setIcon(gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]->icon());
    LightGroupAct->setStatusTip(gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]->statusTip());
    LightGroupAct->setProperty("CommandId", QVariant(LC_EDIT_ACTION_LIGHT));

    gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]->setProperty("CommandId", QVariant(LC_EDIT_ACTION_LIGHT));
    gMainWindow->mActions[LC_EDIT_ACTION_SUNLIGHT]->setProperty("CommandId", QVariant(LC_EDIT_ACTION_SUNLIGHT));
    gMainWindow->mActions[LC_EDIT_ACTION_SPOTLIGHT]->setProperty("CommandId", QVariant(LC_EDIT_ACTION_SPOTLIGHT));
    gMainWindow->mActions[LC_EDIT_ACTION_AREALIGHT]->setProperty("CommandId", QVariant(LC_EDIT_ACTION_AREALIGHT));

    gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]->setProperty("CommandTip", QVariant("Lights - Pointlight"));
    gMainWindow->mActions[LC_EDIT_ACTION_SUNLIGHT]->setProperty("CommandTip", QVariant("Lights - Sunlight"));
    gMainWindow->mActions[LC_EDIT_ACTION_SPOTLIGHT]->setProperty("CommandTip", QVariant("Lights - Spotlight"));
    gMainWindow->mActions[LC_EDIT_ACTION_AREALIGHT]->setProperty("CommandTip", QVariant("Lights - Arealight"));
    connect(LightGroupAct, SIGNAL(triggered()),
            this,          SLOT(  groupActionTriggered()));

    connect(gMainWindow->mActions[LC_EDIT_ACTION_LIGHT],     SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_EDIT_ACTION_SUNLIGHT],  SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_EDIT_ACTION_SPOTLIGHT], SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_EDIT_ACTION_AREALIGHT], SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    // Viewpoint icons
    ViewpointGroupAct = new QAction(tr("Viewpoints"), this);
    ViewpointGroupAct->setObjectName("ViewpointGroupAct.4");
    ViewpointGroupAct->setToolTip(tr("Viewpoints - Home"));
    ViewpointGroupAct->setIcon(gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]->icon());
    ViewpointGroupAct->setStatusTip(gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]->statusTip());
    ViewpointGroupAct->setProperty("CommandId", QVariant(LC_VIEW_VIEWPOINT_HOME));
    lpub->actions.insert(ViewpointGroupAct->objectName(), Action(QStringLiteral("3DViewer.Viewpoints"), ViewpointGroupAct));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_FRONT]->setIcon(QIcon(":/resources/front.png"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BACK]->setIcon(QIcon(":/resources/back.png"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_LEFT]->setIcon(QIcon(":/resources/left.png"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_RIGHT]->setIcon(QIcon(":/resources/right.png"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_TOP]->setIcon(QIcon(":/resources/top.png"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BOTTOM]->setIcon(QIcon(":/resources/bottom.png"));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_FRONT]->setProperty("CommandId", QVariant(LC_VIEW_VIEWPOINT_FRONT));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BACK]->setProperty("CommandId", QVariant(LC_VIEW_VIEWPOINT_BACK));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_LEFT]->setProperty("CommandId", QVariant(LC_VIEW_VIEWPOINT_LEFT));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_RIGHT]->setProperty("CommandId", QVariant(LC_VIEW_VIEWPOINT_RIGHT));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_TOP]->setProperty("CommandId", QVariant(LC_VIEW_VIEWPOINT_TOP));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BOTTOM]->setProperty("CommandId", QVariant(LC_VIEW_VIEWPOINT_BOTTOM));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]->setProperty("CommandId", QVariant(LC_VIEW_VIEWPOINT_HOME));

    gMainWindow->mActions[LC_VIEW_VIEWPOINT_FRONT]-> setProperty("CommandTip", QVariant("Viewpoints - Front"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BACK]->  setProperty("CommandTip", QVariant("Viewpoints - Back"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_LEFT]->  setProperty("CommandTip", QVariant("Viewpoints - Left"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_RIGHT]-> setProperty("CommandTip", QVariant("Viewpoints - Right"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_TOP]->   setProperty("CommandTip", QVariant("Viewpoints - Top"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BOTTOM]->setProperty("CommandTip", QVariant("Viewpoints - Bottom"));
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]->  setProperty("CommandTip", QVariant("Viewpoints - Home"));
    connect(ViewpointGroupAct,                               SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_VIEW_VIEWPOINT_FRONT],  SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_VIEW_VIEWPOINT_BACK],   SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_VIEW_VIEWPOINT_LEFT],   SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_VIEW_VIEWPOINT_RIGHT],  SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_VIEW_VIEWPOINT_TOP],    SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_VIEW_VIEWPOINT_BOTTOM], SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    connect(gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME],   SIGNAL(triggered()),
            this,                                            SLOT(groupActionTriggered()));

    ViewpointZoomExtAct = new QAction(tr("Fit To View"),this);
    ViewpointZoomExtAct->setObjectName("ViewpointZoomExtAct.4");
    ViewpointZoomExtAct->setStatusTip(tr("Specify whether to zoom all items to fit the current view"));
    ViewpointZoomExtAct->setCheckable(true);
    ViewpointZoomExtAct->setChecked(lcGetProfileInt(LC_PROFILE_VIEWPOINT_ZOOM_EXTENT));
    lpub->actions.insert(ViewpointZoomExtAct->objectName(), Action(QStringLiteral("3DViewer.Viewpoints.Fit To View"), ViewpointZoomExtAct));
    connect(ViewpointZoomExtAct, SIGNAL(triggered()), this, SLOT(groupActionTriggered()));

    viewerZoomSliderAct = new QWidgetAction(nullptr);
    viewerZoomSliderWidget = new QSlider();
    viewerZoomSliderWidget->setSingleStep(1);
    viewerZoomSliderWidget->setTickInterval(10);
    viewerZoomSliderWidget->setTickPosition(QSlider::TicksBelow);
    viewerZoomSliderWidget->setMaximum(150);
    viewerZoomSliderWidget->setMinimum(1);
    viewerZoomSliderWidget->setValue(50);
    connect(viewerZoomSliderWidget, SIGNAL(valueChanged(int)), this, SLOT(ViewerZoomSlider(int)));

    gMainWindow->mActions[LC_PIECE_DELETE]->setIcon(QIcon(":/resources/delete.png"));

    gMainWindow->mActions[LC_FILE_SAVE_IMAGE]->setIcon(QIcon(":/resources/saveimage.png"));
    gMainWindow->mActions[LC_FILE_SAVE_IMAGE]->setShortcut(QStringLiteral("Alt+0"));
    gMainWindow->mActions[LC_FILE_SAVE_IMAGE]->setStatusTip(tr("Save an image of the current view - Alt+0"));

    gMainWindow->mActions[LC_FILE_EXPORT_3DS]->setIcon(QIcon(":/resources/3ds32.png"));
    gMainWindow->mActions[LC_FILE_EXPORT_COLLADA]->setIcon(QIcon(":/resources/dae32.png"));
    gMainWindow->mActions[LC_FILE_EXPORT_WAVEFRONT]->setIcon(QIcon(":/resources/obj32.png"));

    QIcon ViewZoomExtentsIcon;
    ViewZoomExtentsIcon.addFile(":/resources/viewzoomextents.png");
    ViewZoomExtentsIcon.addFile(":/resources/viewzoomextents16.png");
    gMainWindow->mActions[LC_VIEW_ZOOM_EXTENTS]->setIcon(ViewZoomExtentsIcon);

    QIcon ViewLookAtIcon;
    ViewLookAtIcon.addFile(":/resources/lookat.png");
    ViewLookAtIcon.addFile(":/resources/lookat.png");
    gMainWindow->mActions[LC_VIEW_LOOK_AT]->setIcon(ViewLookAtIcon);
    gMainWindow->mActions[LC_VIEW_LOOK_AT]->setShortcut(QStringLiteral("Shift+O"));
    gMainWindow->mActions[LC_VIEW_LOOK_AT]->setStatusTip(tr("Position camera so selection is placed at the viewport center - Shift+O"));

    enableBuildModification();
}

void Gui::create3DMenus()
{
     /*
      * Not used
     FileMenuViewer = menuBar()->addMenu(tr("&Step"));
     FileMenuViewer->addAction(gMainWindow->mActions[LC_FILE_SAVEAS]);
     */

     ViewerExportMenu = new QMenu(tr("&Export As..."), this);
     ViewerExportMenu->setIcon(QIcon(":/resources/exportas.png"));
     ViewerExportMenu->addAction(gMainWindow->mActions[LC_FILE_EXPORT_3DS]);
     ViewerExportMenu->addAction(gMainWindow->mActions[LC_FILE_EXPORT_COLLADA]);
     ViewerExportMenu->addAction(gMainWindow->mActions[LC_FILE_EXPORT_WAVEFRONT]);

     SnapXYMenu = new QMenu(tr("Snap XY"), this);
     for (int actionIdx = LC_EDIT_SNAP_MOVE_XY0; actionIdx <= LC_EDIT_SNAP_MOVE_XY9; actionIdx++)
         SnapXYMenu->addAction(gMainWindow->mActions[actionIdx]);

     SnapZMenu = new QMenu(tr("Snap Z"), this);
     for (int actionIdx = LC_EDIT_SNAP_MOVE_Z0; actionIdx <= LC_EDIT_SNAP_MOVE_Z9; actionIdx++)
         SnapZMenu->addAction(gMainWindow->mActions[actionIdx]);

     gMainWindow->mActions[LC_EDIT_ACTION_SELECT]->setMenu(gMainWindow->GetSelectionModeMenu());

     TransformAct->setMenu(gMainWindow->GetTransformMenu());

     SnapMenu = new QMenu(tr("Snap Menu"), this);
     SnapMenu->addAction(gMainWindow->mActions[LC_EDIT_SNAP_MOVE_TOGGLE]);
     SnapMenu->addSeparator();
     SnapMenu->addMenu(SnapXYMenu);
     SnapMenu->addMenu(SnapZMenu);
     MoveAct->setMenu(SnapMenu);

     SnapAngleMenu = new QMenu(tr("Snap Angle Menu"), this);
     SnapAngleMenu->addAction(gMainWindow->mActions[LC_EDIT_SNAP_ANGLE_TOGGLE]);
     SnapAngleMenu->addSeparator();
     for (int actionIdx = LC_EDIT_SNAP_ANGLE0; actionIdx <= LC_EDIT_SNAP_ANGLE9; actionIdx++)
         SnapAngleMenu->addAction(gMainWindow->mActions[actionIdx]);
     AngleAct->setMenu(SnapAngleMenu);

     BuildModMenu = new QMenu(tr("Build Modification"),this);
     BuildModMenu->addAction(UpdateBuildModAct);
     BuildModMenu->addSeparator();
     BuildModMenu->addAction(ApplyBuildModAct);
     BuildModMenu->addAction(RemoveBuildModAct);
     BuildModMenu->addSeparator();
     BuildModMenu->addAction(LoadBuildModAct);
     BuildModMenu->addAction(DeleteBuildModAct);
     CreateBuildModAct->setMenu(BuildModMenu);

     LightMenu = new QMenu(tr("Lights"), this);
     LightMenu->addAction(ApplyLightAct);
     LightMenu->addSeparator();
     LightMenu->addAction(gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]);
     LightMenu->addAction(gMainWindow->mActions[LC_EDIT_ACTION_SUNLIGHT]);
     LightMenu->addAction(gMainWindow->mActions[LC_EDIT_ACTION_SPOTLIGHT]);
     LightMenu->addAction(gMainWindow->mActions[LC_EDIT_ACTION_AREALIGHT]);
     LightGroupAct->setMenu(LightMenu);

     ViewpointGroupAct->setMenu(gMainWindow->GetViewpointMenu());
     gMainWindow->GetViewpointMenu()->insertAction(gMainWindow->mActions[LC_VIEW_VIEWPOINT_FRONT], ViewpointZoomExtAct);
     gMainWindow->GetViewpointMenu()->insertSeparator(gMainWindow->mActions[LC_VIEW_VIEWPOINT_FRONT]);

     // Viewer menus
     ViewerMenu = menuBar()->addMenu(tr("&Visual Editor"));
     // Visual Editor Preferences menu
     ViewerMenu->addAction(gMainWindow->mActions[LC_VIEW_PREFERENCES]);
     ViewerMenu->addSeparator();
     // Render menu
     ViewerMenu->addAction(blenderRenderAct);
     ViewerMenu->addAction(blenderImportAct);
     ViewerMenu->addAction(povrayRenderAct);
     ViewerMenu->addSeparator();
     // Save Image menu
     ViewerMenu->addAction(gMainWindow->mActions[LC_FILE_SAVE_IMAGE]);
     // Export As menu
     ViewerMenu->addMenu(ViewerExportMenu);
     ViewerMenu->addSeparator();
     // Submodel Edit menus
     ViewerMenu->addAction(gMainWindow->mActions[LC_EDIT_CUT]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_EDIT_COPY]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_EDIT_PASTE]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_PIECE_DELETE]);
     ViewerMenu->addSeparator();
     ViewerMenu->addAction(gMainWindow->mActions[LC_PIECE_PAINT_SELECTED]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_PIECE_EDIT_SELECTED_SUBMODEL]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_PIECE_EDIT_END_SUBMODEL]);
     ViewerMenu->addSeparator();
     // Show/Hide Piece menus
     ViewerMenu->addAction(gMainWindow->mActions[LC_PIECE_HIDE_SELECTED]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_PIECE_UNHIDE_SELECTED]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_PIECE_HIDE_UNSELECTED]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_PIECE_UNHIDE_ALL]);
     ViewerMenu->addSeparator();
     // Camera menu
     ViewerMenu->addMenu(gMainWindow->GetCameraMenu());
     // Tools menu
     gMainWindow->GetToolsMenu()->addSeparator();
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_SELECT]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ROTATE]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_MOVE]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_DELETE]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_PAINT]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_COLOR_PICKER]);
     gMainWindow->GetToolsMenu()->addSeparator();
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ROTATESTEP]);
     gMainWindow->GetToolsMenu()->addAction(CreateBuildModAct);
     gMainWindow->GetToolsMenu()->addSeparator();
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_INSERT]);
     gMainWindow->GetToolsMenu()->addAction(LightGroupAct);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_CAMERA]);
     gMainWindow->GetToolsMenu()->addSeparator();
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_VIEW_LOOK_AT]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_VIEW_ZOOM_EXTENTS]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ZOOM]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ZOOM_REGION]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ROTATE_VIEW]);
     gMainWindow->GetToolsMenu()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_PAN]);
     gMainWindow->GetToolsMenu()->addAction(ViewpointGroupAct);
     gMainWindow->GetToolsMenu()->addSeparator();
     gMainWindow->GetToolsMenu()->addAction(TransformAct);
     gMainWindow->GetToolsMenu()->addAction(MoveAct);
     gMainWindow->GetToolsMenu()->addAction(AngleAct); // Snap Rotations to Fixed Intervals menu item
     ViewerMenu->addMenu(gMainWindow->GetToolsMenu());
     // ViewPoint menu
     ViewerMenu->addMenu(gMainWindow->GetViewpointMenu());
     // Projection menu
     ViewerMenu->addMenu(gMainWindow->GetProjectionMenu());
     // Shading menu
     ViewerMenu->addMenu(gMainWindow->GetShadingMenu());
     ViewerMenu->addSeparator();
     // View menus
     ViewerMenu->addAction(gMainWindow->mActions[LC_VIEW_SPLIT_HORIZONTAL]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_VIEW_SPLIT_VERTICAL]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_VIEW_REMOVE_VIEW]);
     ViewerMenu->addAction(gMainWindow->mActions[LC_VIEW_RESET_VIEWS]);
     ViewerMenu->addSeparator();

     // Zoom slider menu item
     connect(gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME], SIGNAL(triggered()), this, SLOT(ResetViewerZoomSlider()));
     viewerZoomSliderAct->setDefaultWidget(viewerZoomSliderWidget);
     ViewerZoomSliderMenu = new QMenu(tr("Zoom Slider"),this);
     ViewerZoomSliderMenu->addAction(viewerZoomSliderAct);
     gMainWindow->mActions[LC_EDIT_ACTION_ZOOM]->setMenu(ViewerZoomSliderMenu);
     ViewerMenu->addSeparator();
     // Visual Editor about menu
     ViewerMenu->addAction(gMainWindow->mActions[LC_HELP_ABOUT]);

     RotateActionMenu = new QMenu(tr("Edit Rotate"),this);
     RotateActionMenu->addAction(EnableRotstepRotateAct);
     RotateActionMenu->addAction(EnableBuildModAct);
     gMainWindow->mActions[LC_EDIT_ACTION_ROTATE]->setMenu(RotateActionMenu);

     CameraMenu = new QMenu(tr("Camera Settings"),this);
     CameraMenu->addAction(ApplyCameraAct);
     CameraMenu->addSeparator();
     CameraMenu->addMenu(gMainWindow->GetCameraMenu());
     CameraMenu->addSeparator();
     CameraMenu->addAction(UseImageSizeAct);
     CameraMenu->addAction(AutoCenterSelectionAct);
     CameraMenu->addAction(DefaultCameraPropertiesAct);
     gMainWindow->mActions[LC_EDIT_ACTION_CAMERA]->setMenu(CameraMenu);
}

void Gui::create3DToolBars()
{
    exportToolBar->addSeparator();
    exportToolBar->addAction(blenderRenderAct);
    exportToolBar->addAction(blenderImportAct);
    exportToolBar->addAction(povrayRenderAct);
    exportToolBar->addSeparator();
    exportToolBar->addAction(gMainWindow->mActions[LC_FILE_SAVE_IMAGE]);
    exportToolBar->addAction(gMainWindow->mActions[LC_FILE_EXPORT_3DS]);
    exportToolBar->addAction(gMainWindow->mActions[LC_FILE_EXPORT_COLLADA]);
    exportToolBar->addAction(gMainWindow->mActions[LC_FILE_EXPORT_WAVEFRONT]);

    gMainWindow->GetToolsToolBar()->addSeparator();
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_SELECT]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ROTATE]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_MOVE]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_DELETE]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_PAINT]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_COLOR_PICKER]);
    gMainWindow->GetToolsToolBar()->addSeparator();
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ROTATESTEP]);
    gMainWindow->GetToolsToolBar()->addAction(CreateBuildModAct);
    gMainWindow->GetToolsToolBar()->addSeparator();
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_INSERT]);
    gMainWindow->GetToolsToolBar()->addAction(LightGroupAct);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_CAMERA]);
    gMainWindow->GetToolsToolBar()->addSeparator();
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_VIEW_LOOK_AT]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_VIEW_ZOOM_EXTENTS]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ZOOM]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ZOOM_REGION]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_ROTATE_VIEW]);
    gMainWindow->GetToolsToolBar()->addAction(gMainWindow->mActions[LC_EDIT_ACTION_PAN]);
    gMainWindow->GetToolsToolBar()->addAction(ViewpointGroupAct);
    gMainWindow->GetToolsToolBar()->addSeparator();
    gMainWindow->GetToolsToolBar()->addAction(TransformAct);
    gMainWindow->GetToolsToolBar()->addAction(MoveAct);
    gMainWindow->GetToolsToolBar()->addAction(AngleAct); // Snap Rotations to Fixed Intervals menu item
    gMainWindow->GetPartsToolBar()->setWindowTitle("Tools Toolbar");
}

bool Gui::eventFilter(QObject *object, QEvent *event)
{
    if ( object == gMainWindow)
        if (event->type() == QEvent::Show || event->type() == QEvent::Hide)
            emit visualEditorVisibleSig(gMainWindow->isVisible());
    return QMainWindow::eventFilter(object, event);
}

void Gui::initiaizeVisualEditor()
{
    connect(this,        SIGNAL(clearViewerWindowSig()),
            gMainWindow, SLOT(  NewProject()));

    connect(this,        SIGNAL(setSelectedPiecesSig(QVector<int>&)),
            gMainWindow, SLOT(  SetSelectedPieces(QVector<int>&)));

    connect(gMainWindow, SIGNAL(SetRotStepMeta()),
            this,        SLOT(  SetRotStepMeta()));

    connect(gMainWindow, SIGNAL(SetRotStepAngleX(float,bool)),
            this,        SLOT(  SetRotStepAngleX(float,bool)));

    connect(gMainWindow, SIGNAL(SetRotStepAngleY(float,bool)),
            this,        SLOT(  SetRotStepAngleY(float,bool)));

    connect(gMainWindow, SIGNAL(SetRotStepAngleZ(float,bool)),
            this,        SLOT(  SetRotStepAngleZ(float,bool)));

    connect(gMainWindow, SIGNAL(SetRotStepTransform(QString&,bool)),
            this,        SLOT(  SetRotStepTransform(QString&,bool)));

    connect(gMainWindow, SIGNAL(GetRotStepMeta()),
            this,        SLOT(  GetRotStepMeta()));

    connect(gMainWindow, SIGNAL(SetActiveModelSig(const QString&,bool)),
            this,        SLOT(  SetActiveModel(const QString&,bool)));

    connect(gMainWindow, SIGNAL(SelectedPartLinesSig(QVector<TypeLine>&,PartSource)),
            this,        SLOT(  SelectedPartLines(QVector<TypeLine>&,PartSource)));

    connect(gMainWindow, SIGNAL(UpdateUndoRedoSig(const QString&,const QString&)),
            this,        SLOT(  UpdateViewerUndoRedo(const QString&,const QString&)));

    connect(gMainWindow, SIGNAL(TogglePreviewWidgetSig(bool)),
            this,        SLOT(  togglePreviewWidget(bool)));

    enable3DActions(false);

    gMainWindow->installEventFilter(gui);

    emit visualEditorVisibleSig(gMainWindow->isVisible());
}

void Gui::enable3DActions(bool enable)
{
    if (enable)
        enableBuildModMenuAndActions();
    else
        CreateBuildModAct->setEnabled(enable);

    LightGroupAct->setEnabled(enable);
    ViewpointGroupAct->setEnabled(enable);

    blenderRenderAct->setEnabled(enable);
    blenderImportAct->setEnabled(enable);
    povrayRenderAct->setEnabled(enable);
    ViewerExportMenu->setEnabled(enable);

    GetToolsToolBar()->setEnabled(enable);
    GetTimelineToolBar()->setEnabled(enable);
    GetPropertiesToolBar()->setEnabled(enable);
    GetPartsToolBar()->setEnabled(enable);
    GetColorsToolBar()->setEnabled(enable);

    GetCameraMenu()->setEnabled(enable);
    GetToolsMenu()->setEnabled(enable);
    GetViewpointMenu()->setEnabled(enable);
    GetProjectionMenu()->setEnabled(enable);
    GetShadingMenu()->setEnabled(enable);

    //Window
    gMainWindow->mActions[LC_VIEW_SPLIT_HORIZONTAL]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_SPLIT_VERTICAL]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_REMOVE_VIEW]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_RESET_VIEWS]->setEnabled(enable);
    //File
    gMainWindow->mActions[LC_FILE_RENDER]->setEnabled(enable);
    gMainWindow->mActions[LC_FILE_SAVE_IMAGE]->setEnabled(enable);
    gMainWindow->mActions[LC_FILE_SAVE_IMAGE]->setEnabled(enable);
    //Export
    gMainWindow->mActions[LC_FILE_EXPORT_3DS]->setEnabled(enable);
    gMainWindow->mActions[LC_FILE_EXPORT_BRICKLINK]->setEnabled(enable);
    gMainWindow->mActions[LC_FILE_EXPORT_COLLADA]->setEnabled(enable);
    gMainWindow->mActions[LC_FILE_EXPORT_CSV]->setEnabled(enable);
    gMainWindow->mActions[LC_FILE_EXPORT_HTML]->setEnabled(enable);
    gMainWindow->mActions[LC_FILE_EXPORT_POVRAY]->setEnabled(enable);
    gMainWindow->mActions[LC_FILE_EXPORT_WAVEFRONT]->setEnabled(enable);
    //Tools
    //mActions[LC_EDIT_ACTION_ROTATESTEP]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_CAMERA]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_CLEAR_TRANSFORM]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_SELECT]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_ROTATE]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_PAN]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_ROTATE_VIEW]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_ROLL]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_ZOOM_REGION]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_MOVE]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_INSERT]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_PAINT]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_COLOR_PICKER]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_DELETE]->setEnabled(enable);
    //Shading
    gMainWindow->mActions[LC_VIEW_SHADING_WIREFRAME]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_SHADING_FLAT]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_SHADING_DEFAULT_LIGHTS]->setEnabled(enable);
    //View
    gMainWindow->mActions[LC_VIEW_ZOOM_EXTENTS]->setEnabled(enable);
    gMainWindow->mActions[LC_EDIT_ACTION_ZOOM]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_LOOK_AT]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_SPLIT_HORIZONTAL]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_SPLIT_VERTICAL]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_REMOVE_VIEW]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_RESET_VIEWS]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_FRONT]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BACK]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_LEFT]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_RIGHT]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_TOP]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_BOTTOM]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_CAMERA_NONE]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_PROJECTION_PERSPECTIVE]->setEnabled(enable);
    gMainWindow->mActions[LC_VIEW_PROJECTION_ORTHO]->setEnabled(enable);
}

void Gui::halt3DViewer(bool enable)
{
    if (!gMainWindow)
        return;

    if(enable){
        GetToolsToolBar()->setEnabled(false);
        gMainWindow->menuBar()->setEnabled(false);
    } else {
        GetToolsToolBar()->setEnabled(true);
        gMainWindow->menuBar()->setEnabled(true);
    }

    emit messageSig(LOG_INFO, QString("Exporting content - Visual Editor %1").arg(enable ? "stopped" :"resumed"));
}

void Gui::create3DDockWindows()
{
    //Timeline
    gMainWindow->GetTimelineToolBar()->setWindowTitle(tr("Timeline"));
    gMainWindow->GetTimelineToolBar()->setObjectName("TimelineToolbar");
    gMainWindow->GetTimelineToolBar()->setAllowedAreas(
                Qt::TopDockWidgetArea  | Qt::BottomDockWidgetArea |
                Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
    addDockWidget(Qt::RightDockWidgetArea, gMainWindow->GetTimelineToolBar());
    viewMenu->addAction(gMainWindow->GetTimelineToolBar()->toggleViewAction());

    // split the space covered commandEditDockWindow into two parts, move visualEditDockWindow into the second part.
    splitDockWidget(commandEditDockWindow, gMainWindow->GetTimelineToolBar(), Qt::Vertical);

    // Preview
    if (GetPreferences().mPreviewPosition == lcPreviewPosition::Dockable)
        createPreviewWidget();

    //Properties
    gMainWindow->GetPropertiesToolBar()->setWindowTitle(tr("Properties"));
    gMainWindow->GetPropertiesToolBar()->setObjectName("PropertiesToolbar");
    gMainWindow->GetPropertiesToolBar()->setAllowedAreas(
                Qt::TopDockWidgetArea  | Qt::BottomDockWidgetArea |
                Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
    addDockWidget(Qt::RightDockWidgetArea, gMainWindow->GetPropertiesToolBar());
    viewMenu->addAction(gMainWindow->GetPropertiesToolBar()->toggleViewAction());

    tabifyDockWidget(gMainWindow->GetTimelineToolBar(), gMainWindow->GetPropertiesToolBar());

    //Colors Selection
    gMainWindow->GetColorsToolBar()->setWindowTitle(tr("Colors"));
    gMainWindow->GetColorsToolBar()->setObjectName("ColorsToolbar");
    gMainWindow->GetColorsToolBar()->setAllowedAreas(
                Qt::TopDockWidgetArea  | Qt::BottomDockWidgetArea |
                Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
    addDockWidget(Qt::RightDockWidgetArea, gMainWindow->GetColorsToolBar());
    viewMenu->addAction(gMainWindow->GetColorsToolBar()->toggleViewAction());

    tabifyDockWidget(gMainWindow->GetTimelineToolBar(), gMainWindow->GetColorsToolBar());

    //Part Selection
    gMainWindow->GetPartsToolBar()->setWindowTitle(tr("Parts"));
    gMainWindow->GetPartsToolBar()->setObjectName("PartsToolbar");
    gMainWindow->GetPartsToolBar()->setAllowedAreas(
                Qt::TopDockWidgetArea  | Qt::BottomDockWidgetArea |
                Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
    addDockWidget(Qt::RightDockWidgetArea, gMainWindow->GetPartsToolBar());
    viewMenu->addAction(gMainWindow->GetPartsToolBar()->toggleViewAction());

    tabifyDockWidget(gMainWindow->GetTimelineToolBar(), gMainWindow->GetPartsToolBar());

    //Visual Editor
    visualEditDockWindow = new QDockWidget(tr("Visual Editor"), this);
    visualEditDockWindow->setObjectName("ModelDockWindow");
    visualEditDockWindow->setAllowedAreas(
                Qt::TopDockWidgetArea  | Qt::BottomDockWidgetArea |
                Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
    visualEditDockWindow->setWidget(gMainWindow);
    addDockWidget(Qt::RightDockWidgetArea, visualEditDockWindow);
    viewMenu->addAction(visualEditDockWindow->toggleViewAction());

    tabifyDockWidget(gMainWindow->GetTimelineToolBar(), visualEditDockWindow);

    // Status Bar and Window Flags
    connect(visualEditDockWindow,                SIGNAL (topLevelChanged(bool)), this, SLOT (toggleLCStatusBar(bool)));
    connect(visualEditDockWindow,                SIGNAL (topLevelChanged(bool)), this, SLOT (enableWindowFlags(bool)));
    connect(gMainWindow->GetTimelineToolBar(),   SIGNAL (topLevelChanged(bool)), this, SLOT (enableWindowFlags(bool)));
    connect(gMainWindow->GetPropertiesToolBar(), SIGNAL (topLevelChanged(bool)), this, SLOT (enableWindowFlags(bool)));
    connect(gMainWindow->GetColorsToolBar(),     SIGNAL (topLevelChanged(bool)), this, SLOT (enableWindowFlags(bool)));
    connect(gMainWindow->GetPartsToolBar(),      SIGNAL (topLevelChanged(bool)), this, SLOT (enableWindowFlags(bool)));
}

bool Gui::createPreviewWidget()
{
    gMainWindow->CreatePreviewWidget();

    if (gMainWindow->GetPreviewWidget()) {
        previewDockWindow = new QDockWidget(tr("Preview"), this);
        previewDockWindow->setWindowTitle(tr("Preview"));
        previewDockWindow->setObjectName("PreviewDockWindow");
        previewDockWindow->setAllowedAreas(
                    Qt::TopDockWidgetArea  | Qt::BottomDockWidgetArea |
                    Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
        previewDockWindow->setWidget(gMainWindow->GetPreviewWidget());
        addDockWidget(Qt::RightDockWidgetArea, previewDockWindow);
        viewMenu->addAction(previewDockWindow->toggleViewAction());

        tabifyDockWidget(gMainWindow->GetTimelineToolBar(), previewDockWindow);

        connect(previewDockWindow, SIGNAL (topLevelChanged(bool)), this, SLOT (enableWindowFlags(bool)));

        return true;
    } else {
        messageSig(LOG_ERROR, QString("Preview failed."));
    }
    return false;
}

void Gui::previewPiece(const QString &partType, int colorCode, bool dockable, QRect parentRect, QPoint position)
{
    if (dockable) {
        if (gMainWindow)
            gMainWindow->PreviewPiece(partType, colorCode, false);
        return;
    } else {
        preview = new lcPreview();
        lcViewWidget* viewWidget = new lcViewWidget(nullptr, preview);

        if (preview && viewWidget)
        {
            viewWidget->setAttribute(Qt::WA_DeleteOnClose, true);

            if (preview->SetCurrentPiece(partType, colorCode))
            {
                viewWidget->SetPreviewPosition(parentRect, position);
                return;
            }
        }
    }

    QMessageBox::information(this, tr("Error"), tr("Part preview for '%1' failed.").arg(partType));
}

void Gui::updatePreview()
{
    if (previewDockWindow) {
        gMainWindow->GetPreviewWidget()->UpdatePreview();
    } else if (preview) {
        preview->UpdatePreview();
    }
}

void Gui::togglePreviewWidget(bool visible)
{
    if (previewDockWindow) {
        if (visible)
            previewDockWindow->show();
        else
            previewDockWindow->hide();
        QList<QAction*> viewActions = viewMenu->actions();
        Q_FOREACH (QAction *viewAct, viewActions) {
            if (viewAct->text() == "Preview") {
                viewAct->setChecked(visible);
                viewAct->setVisible(visible);
#ifdef QT_DEBUG_MODE
                emit messageSig(LOG_DEBUG, QString("%1 window %2.")
                                .arg(viewAct->text()).arg(visible ? "Displayed" : "Hidden"));
#endif
                break;
            }
        }
    } else if (visible) {
        createPreviewWidget();
    }
}

void Gui::enableWindowFlags(bool detached)
{
    if (detached) {
        QDockWidget *dockWidget = qobject_cast<QDockWidget *>(sender());
        dockWidget->setWindowFlags(Qt::CustomizeWindowHint |
                                   Qt::Window |
                                   Qt::WindowMinimizeButtonHint |
                                   Qt::WindowMaximizeButtonHint |
                                   Qt::WindowCloseButtonHint);
        dockWidget->show();
    }
}

void Gui::ClearPreviewWidget()
{
    if (gMainWindow && gMainWindow->GetPreviewWidget())
         gMainWindow->GetPreviewWidget()->ClearPreview();
}

void Gui::UpdateViewerUndoRedo(const QString& UndoText, const QString& RedoText)
{
    viewerUndo = !UndoText.isEmpty();
    if (viewerUndo)
    {
        undoAct->setEnabled(true);
        undoAct->setText(QString(tr("&Undo %1")).arg(UndoText));
        undoAct->setStatusTip(tr("Undo %1 - Ctrl+Z").arg(UndoText));
    }
    else
    {
        undoAct->setEnabled(undoStack->canUndo());
        undoAct->setText(tr("&Undo"));
        undoAct->setStatusTip(tr("Undo last change - Ctrl+Z"));
    }

    viewerRedo = !RedoText.isEmpty();
    if (viewerRedo)
    {
        redoAct->setEnabled(true);
        redoAct->setText(QString(tr("&Redo %1")).arg(RedoText));
#ifdef __APPLE__
        redoAct->setStatusTip(tr("Redo %1 - Ctrl+Shift+Z").arg(RedoText));
#else
        redoAct->setStatusTip(tr("Redo %1 - Ctrl+Y").arg(RedoText));
#endif
    }
    else
    {
        redoAct->setEnabled(undoStack->canRedo());
        redoAct->setText(tr("&Redo"));
#ifdef __APPLE__
        redoAct->setStatusTip(tr("Redo last change - Ctrl+Shift+Z"));
#else
        redoAct->setStatusTip(tr("Redo last change - Ctrl+Y"));
#endif
    }
}

void Gui::showLCStatusMessage()
{
    if(!visualEditDockWindow->isFloating())
    statusBarMsg(gMainWindow->mLCStatusBar->currentMessage());
}

void Gui::toggleLCStatusBar(bool topLevel)
{
    Q_UNUSED(topLevel);

    if(visualEditDockWindow->isFloating())
        gMainWindow->statusBar()->show();
    else
        gMainWindow->statusBar()->hide();
}

void Gui::ViewerZoomSlider(int value)
{
    float z = value;
    if (value > mViewerZoomLevel)
        z = 10.0f;
    else
        z = -10.0f;
    gMainWindow->GetActiveView()->Zoom(z);
    mViewerZoomLevel = value;
}

void Gui::ResetViewerZoomSlider()
{
   viewerZoomSliderWidget->setValue(50);
}

void Gui::enableApplyLightAction()
{
    ApplyLightAct->setEnabled(lcGetActiveProject()->GetImageType() == Options::CSI);
}

void Gui::applyLightSettings()
{
    int it = lcGetActiveProject()->GetImageType();
    if (it != Options::CSI)
        return;

    if (lpub->currentStep){

        Meta meta;
        LightData lightData = meta.LeoCad.light.value();
        LightMeta lightMeta = meta.LeoCad.light;
        lightMeta.setValue(lightData);

        lcView* ActiveView = gMainWindow->GetActiveView();

        lcModel* ActiveModel = ActiveView->GetActiveModel();

        QString metaString;
        bool newCommand = true;
        Where top = lpub->currentStep->topOfStep();
        Where bottom = lpub->currentStep->bottomOfStep();

        auto notEqual = [] (const float v1, const float v2)
        {
            return qAbs(v1 - v2) > 0.1f;
        };

        beginMacro("LightSettings");

        // Delete existing LIGHT commands starting at the bottom of the current step
        for (Where walk = bottom - 1; walk >= top.lineNumber; --walk)
            if(readLine(walk).startsWith(lightMeta.preamble))
                deleteLine(walk);

        for (lcLight* Light : ActiveModel->GetLights()) {

            emit messageSig(LOG_INFO, QString("Setting Light [%1]").arg(Light->mName));

            QString Type = "Undefined";
            switch(Light->mLightType)
            {
                case LC_POINTLIGHT:
                    Type = "Point";
                    break;
                case LC_SUNLIGHT:
                    Type = "Sun";
                    break;
                case LC_SPOTLIGHT:
                    Type = "Spot";
                    break;
                case LC_AREALIGHT:
                    Type = "Area";
                    break;
            }

            // Populate existing settings
            QString lightKey = QString("%1 %2").arg(Type).arg(Light->mName);
            if (lpub->currentStep->lightList.contains(lightKey))
                lightMeta.setValue(lpub->currentStep->lightList[lightKey]);

            // Type and Name
            lightMeta.lightType.setValue(Type);
            metaString = lightMeta.lightType.format(false,false);
            metaString.append(QString(" NAME \"%1\"").arg(Light->mName));
            lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);

            // Position
            if (notEqual(Light->mPosition[0], lightData.position.x()) ||
                notEqual(Light->mPosition[1], lightData.position.y()) ||
                notEqual(Light->mPosition[2], lightData.position.z())) {
                lightMeta.position.setValues(Light->mPosition[0],
                                             Light->mPosition[1],
                                             Light->mPosition[2]);
                metaString = lightMeta.position.format(false/*local*/,false/*global*/);
                lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
            }

            // Target Position
            if (notEqual(Light->mTargetPosition[0], lightData.target.x()) ||
                notEqual(Light->mTargetPosition[1], lightData.target.y()) ||
                notEqual(Light->mTargetPosition[2], lightData.target.z())) {
                lightMeta.target.setValues(Light->mTargetPosition[0],
                                           Light->mTargetPosition[1],
                                           Light->mTargetPosition[2]);
                metaString = lightMeta.target.format(false,false);
                lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
            }

            // Colour
            if (notEqual(Light->mLightColor[0], lightData.lightColour.x()) ||
                notEqual(Light->mLightColor[1], lightData.lightColour.y()) ||
                notEqual(Light->mLightColor[2], lightData.lightColour.z())) {
                lightMeta.lightColour.setValues(Light->mLightColor[0],
                                                Light->mLightColor[1],
                                                Light->mLightColor[2]);
                metaString = lightMeta.lightColour.format(false,false);
                lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
            }

            // Specular
            if (notEqual(Light->mLightSpecular, lightData.lightSpecular.value())) {
                lightMeta.lightSpecular.setValue(Light->mLightSpecular);
                metaString = lightMeta.lightSpecular.format(false,false);
                lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
            }

            if (Light->mLightType == LC_SUNLIGHT) {
                // Strength
                if (notEqual(Light->mSpotExponent, lightData.strength.value())) {
                    lightMeta.strength.setValue(Light->mSpotExponent);
                    metaString = lightMeta.strength.format(false,false);
                    lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                }
                // Angle
                if (notEqual(Light->mLightFactor[0], lightData.angle.value())) {
                    lightMeta.angle.setValue(Light->mLightFactor[0]);
                    metaString = lightMeta.angle.format(false,false);
                    lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                }
            } else {
                // Power
                if (notEqual(Light->mSpotExponent, lightData.power.value())) {
                    lightMeta.power.setValue(Light->mSpotExponent);
                    metaString = lightMeta.power.format(false,false);
                    lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                }

                // Cutoff Distance
                if (Light->mEnableCutoff &&
                   (notEqual(Light->mSpotCutoff, lightData.spotCutoff.value()))) {
                    lightMeta.power.setValue(Light->mSpotCutoff);
                    metaString = lightMeta.spotCutoff.format(false,false);
                    lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                }

                switch (Light->mLightType)
                {
                case LC_POINTLIGHT:
                case LC_SPOTLIGHT:
                    // Radius
                    if (notEqual(Light->mLightFactor[0], lightData.radius.value())) {
                        lightMeta.radius.setValue(Light->mLightFactor[0]);
                        metaString = lightMeta.radius.format(false,false);
                        lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                    }
                    if (Light->mLightType == LC_SPOTLIGHT) {
                        // Spot Blend
                        if (notEqual(Light->mLightFactor[1], lightData.spotBlend.value())) {
                            lightMeta.spotBlend.setValue(Light->mLightFactor[1]);
                            metaString = lightMeta.spotBlend.format(false,false);
                            lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                        }
                        // Spot Size
                        if (notEqual(Light->mSpotSize, lightData.spotSize.value())) {
                            lightMeta.spotSize.setValue(Light->mSpotSize);
                            metaString = lightMeta.spotSize.format(false,false);
                            lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                        }
                    }
                    break;
                case LC_AREALIGHT:
                    if (Light->mLightShape == LC_LIGHT_SHAPE_RECTANGLE || Light->mLightShape == LC_LIGHT_SHAPE_ELLIPSE) {
                        // Width and Height
                        if (notEqual(Light->mLightFactor[0], lightData.width.value()) ||
                            notEqual(Light->mLightFactor[1], lightData.height.value())) {
                            lightMeta.width.setValue(Light->mLightFactor[0]);
                            metaString = lightMeta.width.format(false,false);
                            metaString.append(QString(" HEIGHT %1").arg(double(Light->mLightFactor[1])));
                            lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                        }
                    } else {
                        // Size
                        if (notEqual(Light->mLightFactor[0], lightData.size.value())) {
                            lightMeta.size.setValue(Light->mLightFactor[0]);
                            metaString = lightMeta.size.format(false,false);
                            lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                        }
                    }

                    // Shape
                    QString Shape = "Undefined";
                    switch(Light->mLightShape) {
                    case LC_LIGHT_SHAPE_SQUARE:
                        Shape = "Square";
                        break;
                    case LC_LIGHT_SHAPE_DISK:
                        Shape = "Disk";
                        break;
                    case LC_LIGHT_SHAPE_RECTANGLE:
                        Shape = "Rectangle";
                        break;
                    case LC_LIGHT_SHAPE_ELLIPSE:
                        Shape = "Ellipse";
                        break;
                    }
                    if (notEqual(Light->mLightShape, lightData.radius.value())) {
                        lightMeta.lightShape.setValue(Shape);
                        metaString = lightMeta.lightShape.format(false,false);
                        lpub->currentStep->mi(it)->setMetaAlt(top, metaString, newCommand);
                    }
                    break;
                }
            }
        }

        endMacro();
    }
}

void Gui::applyCameraSettings()
{
    if (lpub->currentStep){

        lcView* ActiveView = gMainWindow->GetActiveView();

        if (!ActiveView)
            return;

        SettingsMeta cameraMeta;

        if (AutoCenterSelectionAct->isChecked())
            ActiveView->LookAt();

        lcCamera* Camera = ActiveView->GetCamera();

        auto validCameraFoV = [this, &cameraMeta, &Camera] ()
        {
            float result = qRound(Camera->m_fovy);

            if (Preferences::preferredRenderer != RENDERER_NATIVE)
                  // e.g.            30.0  +                 0.01         - 30.0
                   result = Camera->m_fovy + cameraMeta.cameraFoV.value() - GetPreferences().mCFoV;

            return qRound(result);
        };

        auto notEqual = [] (const float v1, const float v2)
        {
            return qAbs(v1 - v2) > 0.1f;
        };

        emit messageSig(LOG_INFO, QString("Setting %1 Camera").arg(Camera->GetName().isEmpty() ? "Default" : Camera->GetName()));

        QString imageFileName;

        int it = lcGetActiveProject()->GetImageType();
        switch(it){
        case Options::PLI:
            cameraMeta.cameraAngles   = lpub->currentStep->pli.pliMeta.cameraAngles;
            cameraMeta.cameraDistance = lpub->currentStep->pli.pliMeta.cameraDistance;
            cameraMeta.modelScale     = lpub->currentStep->pli.pliMeta.modelScale;
            cameraMeta.cameraFoV      = lpub->currentStep->pli.pliMeta.cameraFoV;
            cameraMeta.cameraZNear    = lpub->currentStep->pli.pliMeta.cameraZNear;
            cameraMeta.cameraZFar     = lpub->currentStep->pli.pliMeta.cameraZFar;
            cameraMeta.isOrtho        = lpub->currentStep->pli.pliMeta.isOrtho;
            cameraMeta.imageSize      = lpub->currentStep->pli.pliMeta.imageSize;
            cameraMeta.target         = lpub->currentStep->pli.pliMeta.target;
            cameraMeta.position       = lpub->currentStep->pli.pliMeta.position;
            cameraMeta.upvector       = lpub->currentStep->pli.pliMeta.upvector;
            break;
        case Options::SMP:
            cameraMeta.cameraAngles   = lpub->currentStep->subModel.subModelMeta.cameraAngles;
            cameraMeta.cameraDistance = lpub->currentStep->subModel.subModelMeta.cameraDistance;
            cameraMeta.modelScale     = lpub->currentStep->subModel.subModelMeta.modelScale;
            cameraMeta.cameraFoV      = lpub->currentStep->subModel.subModelMeta.cameraFoV;
            cameraMeta.cameraZNear    = lpub->currentStep->subModel.subModelMeta.cameraZNear;
            cameraMeta.cameraZFar     = lpub->currentStep->subModel.subModelMeta.cameraZFar;
            cameraMeta.isOrtho        = lpub->currentStep->subModel.subModelMeta.isOrtho;
            cameraMeta.imageSize      = lpub->currentStep->subModel.subModelMeta.imageSize;
            cameraMeta.target         = lpub->currentStep->subModel.subModelMeta.target;
            cameraMeta.position       = lpub->currentStep->subModel.subModelMeta.position;
            cameraMeta.upvector       = lpub->currentStep->subModel.subModelMeta.upvector;
            break;
        default: /*Options::CSI:*/
            cameraMeta                = lpub->currentStep->csiStepMeta;
            imageFileName             = lpub->currentStep->pngName;
            break;
        }

        QString metaString;
        bool newCommand = false;
        bool clearStepCache = false;
        lcVector3 ldrawVector;
        Where undefined = Where();
        Where top = lpub->currentStep->topOfStep();
        Where bottom = lpub->currentStep->bottomOfStep();

        float Latitude, Longitude, Distance;
        Camera->GetAngles(Latitude, Longitude, Distance);

        bool applyTarget = !(Camera->mTargetPosition[0] == 0.0f  &&
                             Camera->mTargetPosition[1] == 0.0f  &&
                             Camera->mTargetPosition[2] == 0.0f);
        bool applyPosition = !(Camera->mPosition[0] == 0.0f &&
                             Camera->mPosition[1] == -1.0f &&
                             Camera->mPosition[2] == 0.0);
        bool applyUpVector = applyPosition && !(Camera->mUpVector[0] == 0.0f &&
                             Camera->mUpVector[1] == 0.0f &&
                             Camera->mUpVector[2] == 1.0);
        bool applyZPlanes = applyUpVector;


        beginMacro("CameraSettings");

        // execute first in last out

        if (applyUpVector) {

            clearStepCache = true;
            if (QFileInfo(imageFileName).exists())
                clearStepCSICache(imageFileName);

            // Camera Globe, Switch Y and Z axis with -Y(LC -Z) in the up direction
            ldrawVector = lcVector3LeoCADToLDraw(Camera->mUpVector);
            cameraMeta.upvector.setValues(ldrawVector[0], ldrawVector[1], ldrawVector[2]);
            metaString = cameraMeta.upvector.format(true/*local*/,false/*global*/);
            newCommand = cameraMeta.upvector.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.upvector.here(), metaString, newCommand);
        }

        if (applyTarget) {

            clearStepCache = true;
            if (QFileInfo(imageFileName).exists())
                clearStepCSICache(imageFileName);

            // Camera Globe, Switch Y and Z axis with -Y(LC -Z) in the up direction
            lcVector3 ldrawTarget = lcVector3LeoCADToLDraw(Camera->mTargetPosition);
            cameraMeta.target.setValues(ldrawTarget[0], ldrawTarget[1], ldrawTarget[2]);
            metaString = cameraMeta.target.format(true/*local*/,false/*global*/);
            newCommand = cameraMeta.target.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.target.here(), metaString, newCommand);
        }

        if (applyPosition) {

            clearStepCache = true;
            if (QFileInfo(imageFileName).exists())
                clearStepCSICache(imageFileName);

            // Camera Globe, Switch Y and Z axis with -Y(LC -Z) in the up direction
            ldrawVector = lcVector3LeoCADToLDraw(Camera->mPosition);
            cameraMeta.position.setValues(ldrawVector[0], ldrawVector[1], ldrawVector[2]);
            metaString = cameraMeta.position.format(true/*local*/,false/*global*/);
            newCommand = cameraMeta.position.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.position.here(), metaString, newCommand);
        }

        if (UseImageSizeAct->isChecked()) {
            cameraMeta.imageSize.setFormats(3,0,"###9");
            cameraMeta.imageSize.setValues(lcGetActiveProject()->GetImageWidth(),
                                           lcGetActiveProject()->GetImageHeight());
            metaString = cameraMeta.imageSize.format(true,false);
            newCommand = cameraMeta.imageSize.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.imageSize.here(), metaString, newCommand);
        }

        if (notEqual(Camera->GetScale(), cameraMeta.modelScale.value())) {
            clearStepCache = true;
            cameraMeta.modelScale.setValue(Camera->GetScale());
            metaString = cameraMeta.modelScale.format(true,false);
            newCommand = cameraMeta.modelScale.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.modelScale.here(), metaString, newCommand);
        }

        if (notEqual(qRound(Distance), cameraMeta.cameraDistance.value()) &&
                    !UseImageSizeAct->isChecked()) {
            clearStepCache = true;
            cameraMeta.cameraDistance.setValue(qRound(Distance));
            metaString = cameraMeta.cameraDistance.format(true,false);
            newCommand = cameraMeta.cameraDistance.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.cameraDistance.here(), metaString, newCommand);
        }

        if (notEqual(qRound(Latitude), cameraMeta.cameraAngles.value(0)) ||
            notEqual(qRound(Longitude),cameraMeta.cameraAngles.value(1))) {
            clearStepCache = true;
            cameraMeta.cameraAngles.setFormats(5,2,"###9");
            cameraMeta.cameraAngles.setValues(qRound(Latitude), qRound(Longitude));
            metaString = cameraMeta.cameraAngles.format(true,false);
            newCommand = cameraMeta.cameraAngles.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.cameraAngles.here(), metaString, newCommand);
        }

        if (applyZPlanes && notEqual(cameraMeta.cameraZNear.value(), Camera->m_zNear)) {
            clearStepCache = true;
            cameraMeta.cameraZNear.setValue(Camera->m_zNear);
            metaString = cameraMeta.cameraZNear.format(true,false);
            newCommand = cameraMeta.cameraZNear.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.cameraZNear.here(), metaString, newCommand);
        }

        if (applyZPlanes && notEqual(cameraMeta.cameraZFar.value(), Camera->m_zFar)) {
            clearStepCache = true;
            cameraMeta.cameraZFar.setValue(Camera->m_zFar);
            metaString = cameraMeta.cameraZFar.format(true,false);
            newCommand = cameraMeta.cameraZFar.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.cameraZFar.here(), metaString, newCommand);
        }

        float fovy = validCameraFoV();
        if (notEqual(cameraMeta.cameraFoV.value(), fovy)) {
            clearStepCache = true;
            cameraMeta.cameraFoV.setValue(fovy);
            metaString = cameraMeta.cameraFoV.format(true,false);
            newCommand = cameraMeta.cameraFoV.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.cameraFoV.here(), metaString, newCommand);
        }

        if (Camera->IsOrtho() != cameraMeta.isOrtho.value()) {
            cameraMeta.isOrtho.setValue(Camera->IsOrtho());
            metaString = cameraMeta.isOrtho.format(true,false);
            newCommand = cameraMeta.isOrtho.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.isOrtho.here(), metaString, newCommand);
        }

        if (!Camera->GetName().isEmpty()) {
            cameraMeta.cameraName.setValue(Camera->GetName());
            metaString = cameraMeta.cameraName.format(true,false);
            newCommand = cameraMeta.cameraName.here() == undefined;
            lpub->currentStep->mi(it)->setMetaAlt(newCommand ? top : cameraMeta.cameraName.here(), metaString, newCommand);
        }

        if (clearStepCache && QFileInfo(imageFileName).exists())
            clearStepCSICache(imageFileName);

        endMacro();
    }
}

bool Gui::loadBanner(const int &type, const QString &banner) {

    const bool bom = type == GENERATE_BOM;
    const bool error = type == ERROR_ENCOUNTERED;
    const QString description = bom ? "BOM" : error ? "Error" : "Print";
    QList<QString> bannerData;
    bannerData << "0 " + description + " Banner";
    bannerData << "0 Name: banner.ldr";
    bannerData << "0 Author: Trevor SANDY";
    bannerData << "0 !LDRAW_ORG Unofficial_Model";
    if (error) {
        bannerData << "0 !LEOCAD GROUP BEGIN Minifig Error";
        bannerData << "1 320 -5.805 -96 -0.8715 0.642788 0 -0.766044 0 1 0 0.766044 0 0.642788 3624.dat";
        bannerData << "1 14 -5.805 -96 -0.8715 0.642788 0 -0.766044 0 1 0 0.766044 0 0.642788 3626bp01.dat";
        bannerData << "1 320 -5.805 -72 -0.8715 0.642788 0 -0.766044 0 1 0 0.766044 0 0.642788 973.dat";
        bannerData << "1 0 -5.805 -40 -0.8715 0.642788 0 -0.766044 0 1 0 0.766044 0 0.642788 3815.dat";
        bannerData << "1 20 4.1916 -63.12 11.042 0.633425 0.618968 -0.464382 -0.170055 0.696808 0.696808 0.754887 -0.362405 0.546634 3819.dat";
        bannerData << "1 20 -15.8017 -63.12 -12.7851 0.633425 -0.109309 -0.766044 0.170055 0.985435 -0 0.754887 -0.13027 0.642788 3818.dat";
        bannerData << "1 14 23.763 -57.699 2.4644 0.639017 -0.069521 -0.766044 0.108156 0.994134 0 0.761551 -0.082852 0.642788 3820.dat";
        bannerData << "1 14 -13.3852 -45.247 -25.4625 0.633425 -0.618968 -0.464382 0.170055 0.696808 -0.696808 0.754887 0.362405 0.546634 3820.dat";
        bannerData << "1 23 39.9124 -90.0903 -4.5972 -0.963225 0.041837 0.265419 -0.260218 0.100936 -0.96026 -0.066965 -0.994013 -0.086337 3900.dat";
        bannerData << "1 0 -5.805 -28 -0.8715 0.642788 0 -0.766044 0 1 0 0.766044 0 0.642788 3817.dat";
        bannerData << "1 0 -5.805 -28 -0.8715 0.642788 0 -0.766044 0 1 0 0.766044 0 0.642788 3816.dat";
        bannerData << "0 !LEOCAD GROUP END";
        bannerData << "1 7 -27.4663 -8 -45.3551 0.642788 0 -0.766044 0 1 0 0.766044 0 0.642788 3570p01.dat";
    } else if (bom) {
        bannerData << "0 !LEOCAD GROUP BEGIN Minifig BOM";
        bannerData << "1 320 30.7866 -24 86.6357 0.906308 0 -0.422618 0 1 0 0.422618 0 0.906308 3624.dat";
        bannerData << "1 14 30.7866 -24 86.6357 0.906308 0 -0.422618 0 1 0 0.422618 0 0.906308 3626bp01.dat";
        bannerData << "1 320 30.7866 0 86.6357 0.906308 0 -0.422618 0 1 0 0.422618 0 0.906308 973.dat";
        bannerData << "1 0 30.7866 32 86.6357 0.906308 0 -0.422618 0 1 0 0.422618 0 0.906308 3815.dat";
        bannerData << "1 20 44.8815 8.88 93.2084 0.893107 0.349773 -0.282876 -0.170055 0.844683 0.507536 0.416463 -0.405179 0.813872 3819.dat";
        bannerData << "1 20 16.6916 8.8801 80.0631 0.893107 -0.154122 -0.422619 0.170055 0.985435 -0 0.416463 -0.071868 0.906308 3818.dat";
        bannerData << "1 14 58.8216 19.0033 79.4534 0.893107 0.047303 -0.447351 -0.170055 0.956163 -0.238398 0.416463 0.288989 0.862 3820.dat";
        bannerData << "1 14 13.5239 26.7531 67.5522 0.893107 -0.407817 -0.189856 0.170055 0.696808 -0.696807 0.416463 0.590037 0.691675 3820.dat";
        bannerData << "1 0 30.7866 44 86.6357 0.906308 0 -0.422618 0 1 0 0.422618 0 0.906308 3817.dat";
        bannerData << "1 0 30.7866 44 86.6357 0.906308 0 -0.422618 0 1 0 0.422618 0 0.906308 3816.dat";
        bannerData << "0 !LEOCAD GROUP END";
        bannerData << "1 71 -26.3683 -24 -56.5914 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005ptb.dat";
        bannerData << "1 71 -19.5279 -24 -37.7975 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005pti.dat";
        bannerData << "1 71 -12.6875 -24 -19.0037 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005ptl.dat";
        bannerData << "1 71 -5.8471 -24 -0.2098 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005ptl.dat";
        bannerData << "1 71 -19.5279 0 -37.7975 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005pto.dat";
        bannerData << "1 71 -12.6875 0 -19.0037 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005ptf.dat";
        bannerData << "1 72 -5.8471 0 -0.2098 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005.dat";
        bannerData << "1 72 -26.3683 0 -56.5914 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005.dat";
        bannerData << "1 71 -40.0492 24 -94.1791 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005ptm.dat";
        bannerData << "1 71 -33.2087 24 -75.3853 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005pta.dat";
        bannerData << "1 71 -26.3683 24 -56.5914 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005ptt.dat";
        bannerData << "1 71 -19.5279 24 -37.7975 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005pte.dat";
        bannerData << "1 71 -12.6875 24 -19.0037 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005ptr.dat";
        bannerData << "1 71 -5.8471 24 -0.2098 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005pti.dat";
        bannerData << "1 71 0.9933 24 18.5841 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005pta.dat";
        bannerData << "1 71 7.8337 24 37.378 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005ptl.dat";
        bannerData << "1 72 -19.5279 -48 -37.7975 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3062a.dat";
        bannerData << "1 72 -12.6875 -48 -19.0037 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3062a.dat";
        bannerData << "1 72 0.9933 0 18.5841 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005.dat";
        bannerData << "1 72 -33.2087 0 -75.3853 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 3005.dat";
        bannerData << "1 72 -16.1077 48 -28.4006 0.34202 0 -0.939693 0 1 0 0.939693 0 0.34202 6111.dat";
    } else {
        bannerData << "0 !LEOCAD GROUP BEGIN Minifig Export";
        bannerData << "1 320 66.3171 -69 30.467 0.97437 0 0.224951 0 1 0 -0.224951 0 0.97437 3624.dat";
        bannerData << "1 14 66.3171 -69 30.467 0.97437 0 0.224951 0 1 0 -0.224951 0 0.97437 3626bp01.dat";
        bannerData << "1 320 66.3171 -45 30.467 0.97437 0 0.224951 0 1 0 -0.224951 0 0.97437 973.dat";
        bannerData << "1 0 66.3171 -13 30.467 0.97437 0 0.224951 0 1 0 -0.224951 0 0.97437 3815.dat";
        bannerData << "1 20 81.4705 -36.1201 26.9685 0.960178 0.165696 0.224951 -0.170055 0.985435 0 -0.221674 -0.038254 0.97437 3819.dat";
        bannerData << "1 20 51.1636 -36.1201 33.9654 0.960178 -0.279308 -0.006704 0.170055 0.565223 0.80722 -0.221674 -0.776215 0.590212 3818.dat";
        bannerData << "1 14 87.1702 -18.2469 15.3895 0.960178 0.276229 0.041899 -0.170055 0.696808 -0.696807 -0.221674 0.661934 0.716034 3820.dat";
        bannerData << "1 14 41.1228 -34.3034 14.4235 0.960178 -0.202241 0.19276 0.170055 0.970464 0.171119 -0.221674 -0.131525 0.96621 3820.dat";
        bannerData << "1 0 66.3171 -1 30.467 0.97437 -0.224951 0 0 0 1 -0.224951 -0.97437 0 3817.dat";
        bannerData << "1 0 66.3171 -1 30.467 0.97437 -0.224951 0 0 0 1 -0.224951 -0.97437 0 3816.dat";
        bannerData << "0 !LEOCAD GROUP END";
        bannerData << "1 71 3.7112 0 -12.0337 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 3020.dat";
        bannerData << "1 71 25.9016 -8 10.4962 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 3024.dat";
        bannerData << "1 71 25.9016 -16 10.4962 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 3024.dat";
        bannerData << "1 71 -27.5591 -8 -16.7433 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 3024.dat";
        bannerData << "1 71 -27.5591 -16 -16.7433 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 3024.dat";
        bannerData << "1 71 -27.5591 -32 -16.7433 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 6091.dat";
        bannerData << "1 71 25.9016 -32 10.4962 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 6091.dat";
        bannerData << "1 71 25.9016 -32 10.4962 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 30039.dat";
        bannerData << "1 2 -27.5591 -32 -16.7433 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 30039.dat";
        bannerData << "1 71 -0.8287 -24 -3.1235 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 3937.dat";
        bannerData << "1 72 8.2511 -8 -20.9438 0.891007 0 -0.45399 0 1 0 0.45399 0 0.891007 3023.dat";
        bannerData << "1 72 8.2511 -8 -20.9438 -0.891007 0 0.45399 0 1 0 -0.45399 0 -0.891007 85984.dat";
        bannerData << "1 71 0.872 -23.272 -6.4615 -0.891007 -0.17025 0.420859 0 0.927021 0.375008 -0.45399 0.334135 -0.825982 3938.dat";
        bannerData << "1 72 4.9537 -45.523 -14.4722 -0.891007 -0.17025 0.420859 0 0.927021 0.375008 -0.45399 0.334135 -0.825982 4865a.dat";
        switch (type) {
        case EXPORT_PNG:
            bannerData << "1 25 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptp.dat";
            bannerData << "1 25 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptn.dat";
            bannerData << "1 25 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptg.dat";
            break;
        case EXPORT_JPG:
            bannerData << "1 92 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptj.dat";
            bannerData << "1 92 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptp.dat";
            bannerData << "1 92 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptg.dat";
            break;
        case EXPORT_BMP:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptb.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptm.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptp.dat";
            break;
        case EXPORT_WAVEFRONT:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bpto.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptb.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptj.dat";
            break;
        case EXPORT_COLLADA:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptd.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bpta.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bpte.dat";
            break;
        case EXPORT_3DS_MAX:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptx.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptd.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bpts.dat";
            break;
        case EXPORT_STL:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bpts.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptt.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptl.dat";
            break;
        case EXPORT_POVRAY:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptp.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bpto.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptv.dat";
            break;
        case EXPORT_HTML_PARTS:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bpth.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptt.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptm.dat";
            break;
        case EXPORT_CSV:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptc.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bpts.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptv.dat";
            break;
        case EXPORT_BRICKLINK:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptb.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptl.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptk.dat";
            break;
        default:
            bannerData << "1 320 -1.3668 -4.0043 -50.5269 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptp.dat";
            bannerData << "1 320 16.4534 -4.0043 -41.447 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptd.dat";
            bannerData << "1 320 34.2736 -4.0043 -32.3671 0.891007 -0.411364 -0.192061 0 0.42305 -0.906106 0.45399 0.807347 0.37694 3070bptf.dat";
        }
    }
    bannerData << "0";

    QFile bannerFile(banner);
    if ( ! bannerFile.open(QIODevice::ReadWrite | QIODevice::Truncate | QIODevice::Text)) {
        emit gui->messageSig(LOG_ERROR,tr("Could not open banner file '%1' for writing:\n%2")
                             .arg(banner)
                             .arg(bannerFile.errorString()));
        return false;
    }
    QTextStream out(&bannerFile);
    for (int i = 0; i < bannerData.size(); i++)
        out << bannerData[i] << lpub_endl;
    bannerFile.close();

    if (!gMainWindow->OpenProject(bannerFile.fileName()))
    {
        emit gui->messageSig(LOG_ERROR, tr("Could not load banner file '%1'.").arg(bannerFile.fileName()));
        return false;
    }

    return true;
}

void Gui::groupActionTriggered()
{
    QAction* Action = static_cast<QAction*>(sender());
    lcCommandId commandId;
    QString commandTip;
    bool ok;

    if (Action == ViewpointZoomExtAct)
    {
        lcSetProfileInt(LC_PROFILE_VIEWPOINT_ZOOM_EXTENT, ViewpointZoomExtAct->isChecked());
    }
    else if (Action == ViewpointGroupAct)
    {
        commandId = lcCommandId(ViewpointGroupAct->property("CommandId").toInt(&ok));
        if (ok)
            gMainWindow->mActions[commandId]->trigger();
    }
    else if (Action == LightGroupAct)
    {
        commandId = lcCommandId(LightGroupAct->property("CommandId").toInt(&ok));
        if (ok)
            gMainWindow->mActions[commandId]->trigger();
    }
    else {
        commandId = lcCommandId(Action->property("CommandId").toInt(&ok));
        commandTip = Action->property("CommandTip").toString();
        if (ok)
        {
            if (commandId == LC_EDIT_ACTION_LIGHT ||
                commandId == LC_EDIT_ACTION_SUNLIGHT ||
                commandId == LC_EDIT_ACTION_SPOTLIGHT ||
                commandId == LC_EDIT_ACTION_AREALIGHT) {
                LightGroupAct->setIcon(gMainWindow->mActions[commandId]->icon());
                LightGroupAct->setToolTip(tr(commandTip.toLatin1()));
                LightGroupAct->setStatusTip(gMainWindow->mActions[commandId]->statusTip());
                LightGroupAct->setProperty("CommandId", QVariant(commandId));
            } else {
                ViewpointGroupAct->setIcon(gMainWindow->mActions[commandId]->icon());
                ViewpointGroupAct->setToolTip(tr(commandTip.toLatin1()));
                ViewpointGroupAct->setStatusTip(gMainWindow->mActions[commandId]->statusTip());
                ViewpointGroupAct->setProperty("CommandId", QVariant(commandId));
            }
        }
    }
}

void Gui::restoreLightAndViewpointDefaults(){
    LightGroupAct->setToolTip(tr("Lights - Pointlight"));
    LightGroupAct->setIcon(gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]->icon());
    LightGroupAct->setStatusTip(gMainWindow->mActions[LC_EDIT_ACTION_LIGHT]->statusTip());
    LightGroupAct->setProperty("CommandId", QVariant(LC_EDIT_ACTION_LIGHT));

    ViewpointGroupAct->setToolTip(tr("Viewpoints - Home"));
    ViewpointGroupAct->setIcon(gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]->icon());
    ViewpointGroupAct->setStatusTip(gMainWindow->mActions[LC_VIEW_VIEWPOINT_HOME]->statusTip());
    ViewpointGroupAct->setProperty("CommandId", QVariant(LC_VIEW_VIEWPOINT_HOME));
}

void Gui::enableBuildModification()
{
    if (sender() == EnableBuildModAct)
        lcSetProfileInt(LC_PROFILE_BUILD_MODIFICATION, Preferences::buildModEnabled && EnableBuildModAct->isChecked());
    else if (sender() == EnableRotstepRotateAct)
        lcSetProfileInt(LC_PROFILE_BUILD_MODIFICATION, !EnableRotstepRotateAct->isChecked());
    else
        lcSetProfileInt(LC_PROFILE_BUILD_MODIFICATION, Preferences::buildModEnabled);

    bool buildModEnabled = Preferences::buildModEnabled    &&
            lcGetProfileInt(LC_PROFILE_BUILD_MODIFICATION) &&
            (mBuildModRange.first() || buildModsCount())    &&
            !curFile.isEmpty();
    QIcon RotateIcon;
    if (buildModEnabled)
        RotateIcon.addFile(":/resources/rotatebuildmod.png");
    else
        RotateIcon.addFile(":/resources/rotaterotstep.png");

    EnableBuildModAct->setChecked(buildModEnabled);
    EnableRotstepRotateAct->setChecked(!buildModEnabled);
    gMainWindow->mActions[LC_EDIT_ACTION_ROTATE]->setIcon(RotateIcon);
    gMainWindow->mActions[LC_EDIT_ACTION_ROTATESTEP]->setEnabled(!buildModEnabled);
    gApplication->mPreferences.mBuildModificationEnabled = lcGetProfileInt(LC_PROFILE_BUILD_MODIFICATION);
}

void Gui::enableBuildModActions()
{
    if (!CreateBuildModAct->isEnabled() || !Preferences::modeGUI || exporting())
        return;

    Rc buildModStep = BuildModNoActionRc;

    int hasMod = buildModsCount();

    if (lpub->currentStep)
        buildModStep = Rc(getBuildModStep(lpub->currentStep->topOfStep()));

    bool oneMod = hasMod == 1;

    bool appliedMod = false, sourceMod = false, removedMod = false;

    bool dlgTitle = ApplyBuildModAct->text().endsWith("...");

    switch (buildModStep)
    {
        case BuildModSourceRc:
            sourceMod = true;
            break;
        case BuildModApplyRc:
            appliedMod = true;
            break;
        case BuildModRemoveRc:
            removedMod = true;
            break;
        case BuildModNoActionRc:
        default:
            break;
    }

    if (oneMod || !dlgTitle) {
        QList<QAction*> modActions; modActions
                << ApplyBuildModAct
                << RemoveBuildModAct
                << LoadBuildModAct
                << DeleteBuildModAct;
        Q_FOREACH(QAction* action, modActions) {
            QString text = action->text();
            if (oneMod && dlgTitle)
                text.chop(3);
            if (!oneMod && !dlgTitle)
                text.append("...");
            action->setText(text);
        }
    }

    ApplyBuildModAct->setEnabled(hasMod && (!appliedMod || removedMod) && !sourceMod);
    RemoveBuildModAct->setEnabled(hasMod && (appliedMod || !removedMod) && !sourceMod);
    UpdateBuildModAct->setEnabled(hasMod && sourceMod);
    LoadBuildModAct->setEnabled(hasMod && !(sourceMod && oneMod));
    DeleteBuildModAct->setEnabled(hasMod);
}

void Gui::enableBuildModMenuAndActions()
{
    if (!curFile.isEmpty() && Preferences::buildModEnabled)
        CreateBuildModAct->setEnabled(mBuildModRange.first() || buildModsCount());
    enableBuildModification();
}

void Gui::showDefaultCameraProperties()
{
  gApplication->mPreferences.mDefaultCameraProperties = DefaultCameraPropertiesAct->isChecked();
  lcSetProfileInt(LC_PROFILE_DEFAULT_CAMERA_PROPERTIES, DefaultCameraPropertiesAct->isChecked());
}

void Gui::useImageSize()
{
  lcSetProfileInt(LC_PROFILE_USE_IMAGE_SIZE, UseImageSizeAct->isChecked());
}

void Gui::autoCenterSelection()
{
  lcSetProfileInt(LC_PROFILE_AUTO_CENTER_SELECTION, AutoCenterSelectionAct->isChecked());
}

void Gui::createStatusBar()
{
  statusBar()->showMessage(tr("Ready"));
  if (Preferences::modeGUI)
      connect(gMainWindow->mLCStatusBar, SIGNAL(messageChanged(QString)), this, SLOT(showLCStatusMessage()));
}

void Gui::readNativeSettings()
{
    if (!gMainWindow)
        return;

    QSettings Settings;
    Settings.beginGroup(MAINWINDOW);
    gMainWindow->PartSelectionWidgetLoadState(Settings);
    Settings.endGroup();
}

void Gui::writeNativeSettings()
{
    if (!gMainWindow)
        return;

    QSettings Settings;
    Settings.beginGroup(MAINWINDOW);
    gMainWindow->PartSelectionWidgetSaveState(Settings);
    Settings.endGroup();

    gApplication->SaveTabLayout();
}

void Gui::SetActiveModel(const QString &modelName, bool setActive)
{
    if (modelName == VIEWER_MODEL_DEFAULT)
        return;
    if (lcGetActiveProject()->GetImageType() != Options::CSI)
        return;
    if (getCurrentStep() && modelName == getCurrentStep()->topOfStep().modelName)
        return;

    bool displayModelFile = false;
    if (setActive) {
        if (isSubmodel(modelName)) {
            const QString stepKey = getViewerStepKeyWhere(Where(getSubmodelIndex(modelName), 0));
            if (!stepKey.isEmpty()) {
                lpub->setCurrentStep(stepKey);
                displayModelFile = true;
            }
        } else if (getCurrentStep()) {
            displayModelFile = true;
        } else {
            emit messageSig(LOG_ERROR, QString("Active model '%1' not found").arg(modelName));
        }
        if (displayModelFile)
            displayFile(&lpub->ldrawFile, getCurrentStep()->topOfStep());
    }
}

int Gui::GetImageWidth()
{
    return lcGetActiveProject()->GetImageWidth();
}

int Gui::GetImageHeight()
{
    return lcGetActiveProject()->GetImageHeight();
}

void Gui::saveCurrent3DViewerModel(const QString &modelFile)
{
    lcView* ActiveView   = gMainWindow->GetActiveView();
    lcModel* ActiveModel = ActiveView->GetActiveModel();

    if (ActiveModel){
        // Create a copy of the current camera and add it to cameras
        lcCamera* Camera = ActiveView->GetCamera();
        Camera->CreateName(ActiveModel->GetCameras());
        Camera->SetSelected(true);
        ActiveModel->AddCamera(ActiveView->GetCamera());

        // Get the created camera name
        const QString cameraName = QString("Camera %1").arg(ActiveModel->GetCameras().GetSize());

        // Set the created camera
        ActiveView->SetCamera(cameraName);

        // Save the current model
        if (!lcGetActiveProject()->Save(modelFile))
            emit messageSig(LOG_ERROR, QString("Failed to save current model to file [%1]").arg(modelFile));

        // Reset the camera
        bool RemovedCamera = false;
        for (int CameraIdx = 0; CameraIdx < ActiveModel->GetCameras().GetSize(); )
        {
            QString Name = ActiveModel->GetCameras()[CameraIdx]->GetName();
            if (Name == cameraName)
            {
                RemovedCamera = true;
                ActiveModel->RemoveCameraIndex(CameraIdx);
            }
            else
                CameraIdx++;
        }

        ActiveView->SetCamera(Camera, true);
        ActiveView->GetCamera()->SetName(QString());

        if (RemovedCamera)
            gMainWindow->UpdateCameraMenu();

        Camera = nullptr;
    }
}

/*********************************************
 *
 * RotStep Meta
 *
 ********************************************/

void Gui::SetRotStepMeta()
{
    mStepRotation[0] = mRotStepAngleX;
    mStepRotation[1] = mRotStepAngleY;
    mStepRotation[2] = mRotStepAngleZ;

    if (getCurFile() != "") {
        ShowStepRotationStatus();
        Step *currentStep = Gui::getCurrentStep();

        if (lpub->currentStep){
            bool newCommand = currentStep->rotStepMeta.here() == Where();
            int it = lcGetActiveProject()->GetImageType();
            Where top = currentStep->topOfStep();

            RotStepData rotStepData = lpub->currentStep->rotStepMeta.value();
            rotStepData.type    = mRotStepTransform;
            rotStepData.rots[0] = double(mStepRotation[0]);
            rotStepData.rots[1] = double(mStepRotation[1]);
            rotStepData.rots[2] = double(mStepRotation[2]);
            currentStep->rotStepMeta.setValue(rotStepData);
            QString metaString = lpub->currentStep->rotStepMeta.format(false/*no LOCAL tag*/,false);

            if (newCommand){
                if (top.modelName == gui->topLevelFile())
                    currentStep->mi(it)->scanPastLPubMeta(top);

                QString line = gui->readLine(top);
                Rc rc = lpub->page.meta.parse(line,top);
                if (rc == RotStepRc || rc == StepRc){
                   currentStep->mi(it)->replaceMeta(top, metaString);
                } else {
                   currentStep->mi(it)->insertMeta(top, metaString);
                }
            } else {
                currentStep->mi(it)->replaceMeta(top, metaString);
            }
        }
    }
}

void Gui::ShowStepRotationStatus()
{
    QString rotLabel = QString("%1 X: %2 Y: %3 Z: %4 Transform: %5")
                               .arg(lcGetProfileInt(LC_PROFILE_BUILD_MODIFICATION) ? "BUILD MOD ROTATE" : "ROTSTEP")
                               .arg(QString::number(double(mRotStepAngleX), 'f', 2))
                               .arg(QString::number(double(mRotStepAngleY), 'f', 2))
                               .arg(QString::number(double(mRotStepAngleZ), 'f', 2))
                               .arg(mRotStepTransform.toUpper() == "REL" ? "RELATIVE" :
                                    mRotStepTransform.toUpper() == "ABS" ? "ABSOLUTE" : "ADD");
    statusBarMsg(rotLabel);
}

void Gui::loadTheme(){
  Application::instance()->setTheme();
  emit setTextEditHighlighterSig();
  reloadViewer();
}

void Gui::reloadViewer(){
  if (getCurrentStep()){
    Render::LoadViewer(getCurrentStep()->viewerOptions);
  } else {
    Project* NewProject = new Project();
    gApplication->SetProject(NewProject);
    UpdateAllViews();
  }
}

/*********************************************
 *
 * Native viewer convenience calls
 *
 ********************************************/

 lcPiecesLibrary* Gui::GetPiecesLibrary()
 {
     return lcGetPiecesLibrary();
 }


 lcView* Gui::GetActiveView()
 {
     if (gMainWindow)
         return gMainWindow->GetActiveView();
     else
         return new lcView(lcViewType::View, gApplication->mProject->GetMainModel());
 }

 lcModel* Gui::GetActiveModel()
 {
     return gApplication->mProject->GetMainModel();
 }

 Project* Gui::GetActiveProject()
 {
     return gApplication->mProject;
 }

 lcPartSelectionWidget* Gui::GetPartSelectionWidget()
 {
     if (gMainWindow)
         return gMainWindow->GetPartSelectionWidget();
     return nullptr;
 }

 lcPreferences& Gui::GetPreferences()
 {
     return lcGetPreferences();
 }

 bool Gui::GetSubmodelIconsLoaded()
 {
     if (gMainWindow)
         return gMainWindow->mSubmodelIconsLoaded;
     return false;
 }

 void Gui::SetSubmodelIconsLoaded(bool value)
 {
     if (gMainWindow)
         gMainWindow->mSubmodelIconsLoaded = value;
 }

 int Gui::GetLPubStepPieces()
 {
     if (GetActiveModel())
         return GetActiveModel()->GetPieces().GetSize();
     return 0;
 }

 void Gui::UpdateAllViews()
 {
     lcView::UpdateAllViews();
 }

 void Gui::UnloadOfficialPiecesLibrary()
 {
     lcGetPiecesLibrary()->UnloadOfficialLib();
 }

 void Gui::UnloadUnofficialPiecesLibrary()
 {
     lcGetPiecesLibrary()->UnloadUnofficialLib();
 }

 bool Gui::ReloadPiecesLibrary()
 {
     return lcGetPiecesLibrary()->Reload();
 }

 bool Gui::ReloadUnofficialPiecesLibrary()
 {
     return lcGetPiecesLibrary()->ReloadUnofficialLib();
 }

 void Gui::LoadColors()
 {
     lcGetPiecesLibrary()->LoadColors();
 }

 void Gui::LoadDefaults()
 {
     gApplication->mPreferences.LoadDefaults();
 }

 QToolBar* Gui::GetToolsToolBar()
 {
     if (gMainWindow)
         return gMainWindow->GetToolsToolBar();
     return nullptr;
 }

 QDockWidget* Gui::GetTimelineToolBar()
 {
     if (gMainWindow)
         return gMainWindow->GetTimelineToolBar();
     return nullptr;
 }

 QDockWidget* Gui::GetPropertiesToolBar()
 {
     if (gMainWindow)
         return gMainWindow->GetPropertiesToolBar();
     return nullptr;
 }

 QDockWidget* Gui::GetPartsToolBar()
 {
     if (gMainWindow)
         return gMainWindow->GetPartsToolBar();
     return nullptr;
 }

 QDockWidget* Gui::GetColorsToolBar()
 {
     if (gMainWindow)
         return gMainWindow->GetColorsToolBar();
     return nullptr;
 }

 QMenu* Gui::GetCameraMenu()
 {
     if (gMainWindow)
         return gMainWindow->GetCameraMenu();
     return nullptr;
 }

 QMenu* Gui::GetToolsMenu()
 {
     if (gMainWindow)
         return gMainWindow->GetToolsMenu();
     return nullptr;
 }

 QMenu* Gui::GetViewpointMenu()
 {
     if (gMainWindow)
         return gMainWindow->GetViewpointMenu();
     return nullptr;
 }

 QMenu* Gui::GetProjectionMenu()
 {
     if (gMainWindow)
         return gMainWindow->GetProjectionMenu();
     return nullptr;
 }

 QMenu* Gui::GetShadingMenu()
 {
     if (gMainWindow)
         return gMainWindow->GetShadingMenu();
     return nullptr;
 }

/*********************************************
 *
 * build modificaitons
 *
 ********************************************/

void Gui::createBuildModification()
{
    if (!lpub->currentStep || !Preferences::buildModEnabled)
        return;

    if (lcGetActiveProject()->GetImageType() != Options::CSI)
        return;

    lcView* ActiveView = GetActiveView();
    lcModel* ActiveModel = ActiveView->GetActiveModel();

    if (ActiveModel) {
        QString BuildModKey = buildModChangeKey;
        bool edit = ! BuildModKey.isEmpty();
        QString statusLabel = edit ? "Updating" : "Creating";

        emit progressBarPermInitSig();
        emit progressPermRangeSig(0, 0);   // Busy indicator
        emit progressPermMessageSig(QString("%1 Build Modification...").arg(statusLabel));

        if (mBuildModRange.first() || edit){

            emit messageSig(LOG_INFO, QString("%1 BuildMod for Step %2...")
                            .arg(statusLabel)
                            .arg(lpub->currentStep->stepNumber.number));

            // 'load...' default lines from modelFile and 'save...' buildMod lines from Visual Editor
            lcArray<lcGroup*>  mGroups;
            lcArray<lcCamera*> mCameras;
            lcArray<lcLight*>  mLights;
            lcArray<lcPiece*>  mLPubPieces;                     // Pieces in the buildMod - may include removed viewer pieces
            lcArray<lcGroup*>  mLPubGroups;
            lcArray<lcPiece*>  mViewerPieces;                   // All viewer pieces in the step
            lcArray<lcGroup*>  mViewerGroups;
            lcModelProperties  mViewerProperties;
            lcArray<lcPieceControlPoint> ControlPoints;
            QStringList  mLPubFileLines, mViewerFileLines,
                         LPubModContents, ViewerModContents,
                         ModStepKeys;

            bool FadeStep         = Preferences::enableFadeSteps;
            bool HighlightStep    = Preferences::enableHighlightStep && !suppressColourMeta();

            int AddedPieces       = 0;
            int CurrentStep       = 1;
            lcPiece *Piece        = nullptr;
            lcCamera *Camera      = nullptr;
            lcLight  *Light       = nullptr;
            mViewerProperties     = ActiveModel->GetProperties();
            mViewerFileLines      = ActiveModel->GetFileLines();
            mViewerPieces         = ActiveModel->GetPieces();
            lcPiecesLibrary *Library = lcGetPiecesLibrary();

            QString ModStepKey    = lpub->viewerStepKey;
            ModStepKeys           = ModStepKey.split(";");

            // When edit, initialize BuildMod StepPieces, and Begin and End range with the existing values
            int BuildModBegin     = edit ? getBuildModBeginLineNumber(BuildModKey)  : 0;
            int BuildModAction    = edit ? getBuildModActionLineNumber(BuildModKey) : 0;
            int BuildModEnd       = edit ? getBuildModEndLineNumber(BuildModKey)    : 0;

            int ModBeginLineNum   = edit ? BuildModBegin  : mBuildModRange.at(BM_BEGIN_LINE_NUM);
            int ModActionLineNum  = edit ? BuildModAction : mBuildModRange.at(BM_ACTION_LINE_NUM);
            int ModEndLineNum     = edit ? BuildModEnd    : mBuildModRange.at(BM_BEGIN_LINE_NUM);
            int ModStepPieces     = edit ? getBuildModStepPieces(BuildModKey) : 0;    // All pieces in the previous step
            int ModelIndex        = edit ? getSubmodelIndex(getBuildModStepKeyModelName(BuildModKey)) : mBuildModRange.at(BM_MODEL_INDEX);
            int ModStepIndex      = getBuildModStepIndex(lpub->currentStep->topOfStep());
            int ModStepLineNum    = ModStepKeys[BM_STEP_LINE_KEY].toInt();
            int ModStepNum        = ModStepKeys[BM_STEP_NUM_KEY].toInt();
            int ModDisplayPageNum = displayPageNum;
            QString ModelName     = getSubmodelName(ModelIndex);
            buildModChangeKey     = QString();

            if (ModStepKeys[BM_STEP_MODEL_KEY].toInt() != ModelIndex)
                emit messageSig(LOG_ERROR, QString("BuildMod model (%1) '%2' and StepKey model (%3) mismatch")
                                                   .arg(ModelIndex).arg(ModelName).arg(ModStepKeys[BM_STEP_MODEL_KEY]));

            // Check if there is an existing build mod in this Step
            QRegExp lineRx("^0 !LPUB BUILD_MOD BEGIN ");
            if (stepContains(lpub->currentStep->top, lineRx) && !edit) {

                // Get the application icon as a pixmap
                QPixmap _icon = QPixmap(":/icons/lpub96.png");
                if (_icon.isNull())
                    _icon = QPixmap (":/icons/update.png");

                QMessageBox box;
                box.setWindowIcon(QIcon());
                box.setIconPixmap (_icon);
                box.setTextFormat (Qt::RichText);
                box.setWindowFlags (Qt::Dialog | Qt::CustomizeWindowHint | Qt::WindowTitleHint);
                QString title = "<b>" + QMessageBox::tr ("You specified create action for an existing Build Modification") + "</b>";
                QString text = QMessageBox::tr("<br>This action will replace the existing Build Modification."
                                               "<br>You can cancel and select 'Update Build Modification...' from the build modification submenu."
                                               "<br>Do you want to continue with this create action ?");
                box.setText (title);
                box.setInformativeText (text);
                box.setStandardButtons (QMessageBox::Cancel | QMessageBox::Ok);
                box.setDefaultButton   (QMessageBox::Cancel);

                if (box.exec() == QMessageBox::Cancel)
                    return;
            }

            auto GetGroup = [&mGroups](const QString& Name, bool CreateIfMissing)
            {
                lcGroup* nullGroup = nullptr;
                for (lcGroup* Group : mGroups)
                    if (Group->mName == Name)
                        return Group;

                if (CreateIfMissing)
                {
                    lcGroup* Group = new lcGroup();
                    Group->mName = Name;
                    mGroups.Add(Group);

                    return Group;
                }

                return nullGroup;
            };

            auto ConfigurePartLine = [&ModelIndex, &FadeStep, &HighlightStep, &LPubModContents] (int LineTypeIndex, QTextStream &Stream)
            {
                if (!FadeStep && !HighlightStep)
                    return;

                if (!Stream.string()->isEmpty()) {

                    QStringList ModLines = Stream.string()->split(QRegExp("(\\r\\n)|\\r|\\n"), SkipEmptyParts);
                    QString PartLine = ModLines.last();

                    QStringList argv;
                    split(PartLine, argv);

                    if (argv.size() == 15 && argv[0] == "1") {

                        QString NameMod, ColourPrefix;
                        if (FadeStep) {
                            NameMod = FADE_SFX;
                            ColourPrefix = LPUB3D_COLOUR_FADE_PREFIX;
                        } else if (HighlightStep) {
                            NameMod = HIGHLIGHT_SFX;
                            ColourPrefix = LPUB3D_COLOUR_HIGHLIGHT_PREFIX;
                        }

                        // Colour code
                        if (argv[1].startsWith(ColourPrefix)) {
                            int LineNumber;
                            QString LPubPartLine, NewColorCode;
                            bool NewLine = !gui->getSelectedLine(ModelIndex, LineTypeIndex, VIEWER_MOD, LineNumber);

                            if (NewLine) {
                                NewColorCode = argv[1];

                            } else {
                                LPubPartLine = LPubModContents.at(LineTypeIndex);
                                QStringList dargv;
                                split(LPubPartLine, dargv);

                                if (dargv.size() == 15 && dargv[0] == "1") {
                                    NewColorCode = dargv[1];

                                } else {
                                    NewColorCode = argv[1].right(argv[1].size() - ColourPrefix.size());
                                    if (FadeStep && Preferences::fadeStepsUseColour)
                                        NewColorCode = LDRAW_MAIN_MATERIAL_COLOUR;
                                }
                            }

                            argv[1] = NewColorCode;
                        }

                        // fileName
                        QString FileName = QString(argv[argv.size()-1]).toLower();
                        if (FileName.contains(NameMod)) {

                            QString NewFileName = FileName.replace(NameMod,"");
                            argv[argv.size()-1] = NewFileName;
                        }

                        PartLine = argv.join(" ");

                        ModLines.removeLast();
                        ModLines.append(PartLine);

                        Stream.string()->clear();

                        for (int i = 0; i < ModLines.size(); i++)
                            Stream << ModLines.at(i);
                    }
                }
            };

            auto InsertPiece = [] (lcArray<lcPiece*> &mLPubPieces, lcPiece* Piece, int PieceIdx)
            {
                PieceInfo* Info = Piece->mPieceInfo;

                if (!Info->IsModel())
                {
                    lcMesh* Mesh = Info->GetMesh();

                    if (Mesh && Mesh->mVertexCacheOffset == -1)
                        lcGetPiecesLibrary()->mBuffersDirty = true;
                }

                mLPubPieces.InsertAt(PieceIdx, Piece);
            };

            auto AddPiece = [&InsertPiece](lcArray<lcPiece*> &mLPubPieces, lcPiece* Piece)
            {
                for (int PieceIdx = 0; PieceIdx < mLPubPieces.GetSize(); PieceIdx++)
                {
                    if (mLPubPieces[PieceIdx]->GetStepShow() > Piece->GetStepShow())
                    {
                        InsertPiece(mLPubPieces, Piece, PieceIdx);
                        return;
                    }
                }

                InsertPiece(mLPubPieces, Piece, mLPubPieces.GetSize());
            };

            // load LPub content
            QByteArray ByteArray;
            for (int i = 0; i < lpub->ldrawFile.contents(ModelName).size(); i++){
                if (i > ModActionLineNum)
                    break;
                QString ModLine = lpub->ldrawFile.contents(ModelName).at(i);
                if (i >= ModBeginLineNum)
                    LPubModContents.append(ModLine);
                ByteArray.append(ModLine.toUtf8());
                ByteArray.append(QString("\n").toUtf8());
            }
            QBuffer Buffer(&ByteArray);
            Buffer.open(QIODevice::ReadOnly);
            Buffer.seek(0);

            // load mLPubFileLines, mLPubGroups and mLPubPieces, Camera, Lights and LSynth Control Points with LDraw content using QBA buffer
            while (!Buffer.atEnd())
            {
                qint64 Pos = Buffer.pos();
                QString OriginalLine = Buffer.readLine();
                QString Line = OriginalLine.trimmed();
                QTextStream LineStream(&Line, QIODevice::ReadOnly);

                QString Token;
                LineStream >> Token;

                if (Token == QLatin1String("0"))
                {
                    LineStream >> Token;
                    if (Token == QLatin1String("FILE"))
                    {
                        QString Name = LineStream.readAll().trimmed();

                        if (mViewerProperties.mFileName != Name)
                        {
                            Buffer.seek(Pos);
                            break;
                        }

                        continue;
                    }
                    else if (Token == QLatin1String("NOFILE"))
                    {
                        break;
                    }
                    else if (Token == QLatin1String("STEP"))
                    {
                        CurrentStep++;
                        mLPubFileLines.append(OriginalLine);
                        continue;
                    }

                    if (Token != QLatin1String("!LPUB"))
                    {
                        mLPubFileLines.append(OriginalLine);
                        continue;
                    }

                    LineStream >> Token;

                    if (Token == QLatin1String("MODEL"))
                    {
                        mViewerProperties.ParseLDrawLine(LineStream);
                    }
                    else if (Token == QLatin1String("PIECE"))
                    {
                        if (!Piece)
                            Piece = new lcPiece(nullptr);

                        Piece->ParseLDrawLine(LineStream);
                    }
                    else if (Token == QLatin1String("CAMERA"))
                    {
                        if (!Camera)
                            Camera = new lcCamera(false);

                        if (Camera->ParseLDrawLine(LineStream))
                        {
                            Camera->CreateName(mCameras);
                            mCameras.Add(Camera);
                            Camera = nullptr;
                        }
                    }
                    else if (Token == QLatin1String("LIGHT"))
                    {
                        if (!Light)
                            Light = new lcLight(0.0f, 0.0f, 0.0f);

                        if (Light->ParseLDrawLine(LineStream))
                        {
                            Light->CreateName(mLights);
                            mLights.Add(Light);
                            Light = nullptr;
                        }
                    }
                    else if (Token == QLatin1String("GROUP"))
                    {
                        LineStream >> Token;

                        if (Token == QLatin1String("BEGIN"))
                        {
                            QString Name = LineStream.readAll().trimmed();
                            lcGroup* Group = GetGroup(Name, true);
                            if (!mLPubGroups.IsEmpty())
                                Group->mGroup = mLPubGroups[mLPubGroups.GetSize() - 1];
                            else
                                Group->mGroup = nullptr;
                            mLPubGroups.Add(Group);
                        }
                        else if (Token == QLatin1String("END"))
                        {
                            if (!mLPubGroups.IsEmpty())
                                mLPubGroups.RemoveIndex(mLPubGroups.GetSize() - 1);
                        }
                    }
                    else if (Token == QLatin1String("SYNTH"))
                    {
                        LineStream >> Token;

                        if (Token == QLatin1String("BEGIN"))
                        {
                            ControlPoints.RemoveAll();
                        }
                        else if (Token == QLatin1String("END"))
                        {
                            ControlPoints.RemoveAll();
                        }
                        else if (Token == QLatin1String("CONTROL_POINT"))
                        {
                            float Numbers[13];
                            for (int TokenIdx = 0; TokenIdx < 13; TokenIdx++)
                                LineStream >> Numbers[TokenIdx];

                            lcPieceControlPoint& PieceControlPoint = ControlPoints.Add();
                            PieceControlPoint.Transform = lcMatrix44(lcVector4(Numbers[3], Numbers[9], -Numbers[6], 0.0f), lcVector4(Numbers[5], Numbers[11], -Numbers[8], 0.0f),
                                                                     lcVector4(-Numbers[4], -Numbers[10], Numbers[7], 0.0f), lcVector4(Numbers[0], Numbers[2], -Numbers[1], 1.0f));
                            PieceControlPoint.Scale = Numbers[12];
                        }
                    }

                    continue;

                }
                else if (Token == QLatin1String("1"))
                {
                    int ColorCode;
                    LineStream >> ColorCode;
                    float IncludeMatrix[12];

                    for (int TokenIdx = 0; TokenIdx < 12; TokenIdx++)
                        LineStream >> IncludeMatrix[TokenIdx];

                    lcMatrix44 IncludeTransform(lcVector4(IncludeMatrix[3], IncludeMatrix[6], IncludeMatrix[9], 0.0f), lcVector4(IncludeMatrix[4], IncludeMatrix[7], IncludeMatrix[10], 0.0f),
                            lcVector4(IncludeMatrix[5], IncludeMatrix[8], IncludeMatrix[11], 0.0f), lcVector4(IncludeMatrix[0], IncludeMatrix[1], IncludeMatrix[2], 1.0f));

                    QString PartId = LineStream.readAll().trimmed();
                    QByteArray CleanId = PartId.toLatin1().toUpper().replace('\\', '/');

                    if (Library->IsPrimitive(CleanId.constData()))
                    {
                        mLPubFileLines.append(OriginalLine);
                    }
                    else
                    {
                        if (!Piece)
                            Piece = new lcPiece(nullptr);

                        if (!mLPubGroups.IsEmpty())
                            Piece->SetGroup(mLPubGroups[mLPubGroups.GetSize() - 1]);

                        PieceInfo* Info = Library->FindPiece(PartId.toLatin1().constData(), lcGetActiveProject(), true, true);

                        float* Matrix = IncludeTransform;
                        lcMatrix44 Transform(lcVector4(Matrix[0], Matrix[2], -Matrix[1], 0.0f), lcVector4(Matrix[8], Matrix[10], -Matrix[9], 0.0f),
                                lcVector4(-Matrix[4], -Matrix[6], Matrix[5], 0.0f), lcVector4(Matrix[12], Matrix[14], -Matrix[13], 1.0f));

                        Piece->SetFileLine(mLPubFileLines.size());
                        Piece->SetPieceInfo(Info, PartId, false);
                        Piece->Initialize(Transform, quint32(CurrentStep));
                        Piece->SetColorCode(quint32(ColorCode));
                        Piece->SetControlPoints(ControlPoints);
                        AddPiece(mLPubPieces, Piece);

                        Piece = nullptr;
                    }
                }
                else
                    mLPubFileLines.append(OriginalLine);

                Library->WaitForLoadQueue();
                Library->mBuffersDirty = true;
                Library->UnloadUnusedParts();

                delete Piece;
                delete Camera;
            }

            Buffer.close();

            // load Viewer content from mViewerFileLines, mViewerGroups and mViewerPieces, Camera, Lights and LSynth Control Points
            bool NewPiece      = false;
            bool SelectedOnly  = false;
            bool PieceInserted = false;
            bool PieceModified = true;
            QLatin1String LineEnding("\r\n");

            QString ViewerModContentsString;
            QTextStream Stream(&ViewerModContentsString, QIODevice::ReadWrite);

            lcStep Step       = 1;
            int CurrentLine   = ModBeginLineNum/*0*/;
            int AddedSteps    = 0;

            int LineIndex     = NEW_PART;
            int LineNumber    = 0;
            int PieceAdjustment = 0;
            int EndModLineNum = ModActionLineNum;

            // Do we have a difference between the number of LPub pieces and Viewer pieces ?
            // If pieces have been added or removed, we capture the delta in PieceAdjustment
            // and AddedPieces.
            PieceAdjustment = AddedPieces = edit ? mViewerPieces.GetSize() - ModStepPieces : mViewerPieces.GetSize() - getViewerStepPartCount(ModStepKey);

            // Adjust EndModLineNum to accomodate removed pieces
            if (PieceAdjustment < 0)
                EndModLineNum -= PieceAdjustment;

#ifdef QT_DEBUG_MODE
            emit messageSig(LOG_DEBUG, QString("%1Pieces [%2], Viewer Pieces Count [%3], LPub Pieces Count [%4]")
                            .arg(PieceAdjustment == 0 ? "" : PieceAdjustment > 0 ? "Added " : "Removed ")
                            .arg(PieceAdjustment  < 0 ? -PieceAdjustment : PieceAdjustment)
                            .arg(mViewerPieces.GetSize()).arg(ModStepPieces));
#endif

            for (lcPiece* Piece : mViewerPieces)
            {
                LineIndex  = Piece->GetLineTypeIndex();

                // We have a new piece (NewPiece set to true) when the piece LineIndex is -1
                NewPiece   = !getSelectedLine(ModelIndex, LineIndex, VIEWER_MOD, LineNumber);

                // If PieceInserted, we must increment the 'original' line number for the next piece
                if (PieceInserted) {
                    LineNumber   += PieceInserted;
                }

                // Added pieces have a line number of 0 so we must reset the LineNumber to EndModLineNum
                // and set PieceInserted to true so we can increment the line number for the next piece.
                // For each NewPiece, we'll decrement PieceAdjustment.
                if ((PieceInserted = NewPiece)) {
                    PieceAdjustment--;
                    LineNumber = EndModLineNum;
                }

                // Set PieceModified. Only modified pieces are added to ViewerModContents
                PieceModified = LineNumber <= ModActionLineNum || Piece->PieceModified() || NewPiece;

#ifdef QT_DEBUG_MODE
                emit messageSig(LOG_DEBUG, QString("Viewer Piece LineIndex [%1], ID [%2], Name [%3], LineNumber [%4], Modified: [%5]")
                                                   .arg(LineIndex < 0 ? QString::number(LineIndex) + "], Added Piece [#"+QString::number(AddedPieces) : QString::number(LineIndex))
                                                   .arg(Piece->GetID())
                                                   .arg(Piece->GetName())
                                                   .arg(LineNumber)
                                                   .arg(PieceModified ? "Yes" : "No"));
#endif

                // If PieceAdjustment is not 0, we increment EndModLineNum as we process each piece
                if (PieceAdjustment > 0)
                     EndModLineNum = LineNumber + 1;

                // No more modified pieces so we exit the loop
                if (LineNumber > EndModLineNum)
                    break;

                // Process the current piece
                if (LineNumber >= ModBeginLineNum && PieceModified) {

                    // Use LPubFileLines
                    while (LineNumber/*Piece->GetFileLine()*/ > CurrentLine && CurrentLine < mLPubFileLines.size()/*mViewerFileLines.size()*/)
                    {
                        QString Line = mLPubFileLines[CurrentLine]/*mViewerFileLines[CurrentLine]*/;
                        QTextStream LineStream(&Line, QIODevice::ReadOnly);

                        QString Token;
                        LineStream >> Token;
                        bool Skip = false;

                        if (Token == QLatin1String("0"))
                        {
                            LineStream >> Token;

                            if (Token == QLatin1String("STEP"))
                            {
                                if (Piece->GetStepShow() > Step)
                                    Step++;
                                else
                                    Skip = true;
                            }
                        }

                        if (!Skip)
                        {
                            Stream << mLPubFileLines[CurrentLine]/*mViewerFileLines[CurrentLine]*/;
                            if (AddedSteps > 0)
                                AddedSteps--;
                        }
                        CurrentLine++;
                    }

                    // Use Viewer Pieces
                    while (Piece->GetStepShow() > Step)
                    {
                        Stream << QLatin1String("0 STEP\r\n");
                        AddedSteps++;
                        Step++;
                    }

                    lcGroup* PieceGroup = Piece->GetGroup();

                    if (PieceGroup)
                    {
                        if (mViewerGroups.IsEmpty() || (!mViewerGroups.IsEmpty() && PieceGroup != mViewerGroups[mViewerGroups.GetSize() - 1]))
                        {
                            lcArray<lcGroup*> PieceParents;

                            for (lcGroup* Group = PieceGroup; Group; Group = Group->mGroup)
                                PieceParents.InsertAt(0, Group);

                            int FoundParent = -1;

                            while (!mViewerGroups.IsEmpty())
                            {
                                lcGroup* Group = mViewerGroups[mViewerGroups.GetSize() - 1];
                                int Index = PieceParents.FindIndex(Group);

                                if (Index == -1)
                                {
                                    mViewerGroups.RemoveIndex(mViewerGroups.GetSize() - 1);
                                    Stream << QLatin1String("0 !LPUB GROUP END\r\n");
                                }
                                else
                                {
                                    FoundParent = Index;
                                    break;
                                }
                            }

                            for (int ParentIdx = FoundParent + 1; ParentIdx < PieceParents.GetSize(); ParentIdx++)
                            {
                                lcGroup* Group = PieceParents[ParentIdx];
                                mViewerGroups.Add(Group);
                                Stream << QLatin1String("0 !LPUB GROUP BEGIN ") << Group->mName << LineEnding;
                            }
                        }
                    }
                    else
                    {
                        while (mViewerGroups.GetSize())
                        {
                            mViewerGroups.RemoveIndex(mViewerGroups.GetSize() - 1);
                            Stream << QLatin1String("0 !LPUB GROUP END\r\n");
                        }
                    }

                    if (Piece->mPieceInfo->GetSynthInfo())
                    {
                        Stream << QLatin1String("0 !LPUB SYNTH BEGIN\r\n");

                        const lcArray<lcPieceControlPoint>& ControlPoints = Piece->GetControlPoints();
                        for (int ControlPointIdx = 0; ControlPointIdx < ControlPoints.GetSize(); ControlPointIdx++)
                        {
                            const lcPieceControlPoint& ControlPoint = ControlPoints[ControlPointIdx];

                            Stream << QLatin1String("0 !LPUB SYNTH CONTROL_POINT");

                            const float* FloatMatrix = ControlPoint.Transform;
                            float Numbers[13] = { FloatMatrix[12], -FloatMatrix[14], FloatMatrix[13], FloatMatrix[0], -FloatMatrix[8], FloatMatrix[4], -FloatMatrix[2], FloatMatrix[10], -FloatMatrix[6], FloatMatrix[1], -FloatMatrix[9], FloatMatrix[5], ControlPoint.Scale };

                            for (int NumberIdx = 0; NumberIdx < 13; NumberIdx++)
                                Stream << ' ' << lcFormatValue(Numbers[NumberIdx], NumberIdx < 3 ? 4 : 6);

                            Stream << LineEnding;
                        }
                    }

                    Piece->SaveLDraw(Stream);

                    ConfigurePartLine(Piece->GetLineTypeIndex(), Stream);

                    if (Piece->mPieceInfo->GetSynthInfo())
                        Stream << QLatin1String("0 !LPUB SYNTH END\r\n");
                }
            }

            while (CurrentLine < mLPubFileLines.size()/*mViewerFileLines.size()*/)
            {
                QString Line = mLPubFileLines[CurrentLine]/*mViewerFileLines[CurrentLine]*/;
                QTextStream LineStream(&Line, QIODevice::ReadOnly);

                QString Token;
                LineStream >> Token;
                bool Skip = false;

                if (Token == QLatin1String("0"))
                {
                    LineStream >> Token;

                    if (Token == QLatin1String("STEP") && AddedSteps-- > 0)
                        Skip = true;
                }

                if (!Skip)
                    Stream << mLPubFileLines[CurrentLine]/*mViewerFileLines[CurrentLine]*/;
                CurrentLine++;
            }

            while (mViewerGroups.GetSize())
            {
                mViewerGroups.RemoveIndex(mViewerGroups.GetSize() - 1);
                Stream << QLatin1String("0 !LPUB GROUP END\r\n");
            }

            for (lcCamera* Camera : mCameras)
                if (!SelectedOnly || Camera->IsSelected())
                    Camera->SaveLDraw(Stream);

            for (lcLight* Light : mLights)
                if (!SelectedOnly || Light->IsSelected())
                    Light->SaveLDraw(Stream);

            ViewerModContents = QString(ViewerModContentsString).split(QRegExp("(\\r\\n)|\\r|\\n"), SkipEmptyParts);

            int BuildModPieces = ViewerModContents.size();

            if (!edit)
                BuildModKey = QString("%1 Mod %2").arg(ModelName).arg(buildModsCount() + 1);

            // Delete meta commands uses the 'original' BuildMod values
            int SaveModBeginLineNum  = edit ? BuildModBegin  : ModBeginLineNum;
            int SaveModActionLineNum = edit ? BuildModAction : ModActionLineNum;
            int SaveModEndLineNum    = edit ? BuildModEnd    : ModEndLineNum;
            int SaveModPieces        = edit ? SaveModEndLineNum - SaveModActionLineNum - 1/*Meta Line*/
                                            : BuildModPieces - (AddedPieces > 0 ? AddedPieces : 0);

#ifdef QT_DEBUG_MODE
            emit messageSig(LOG_TRACE, QString("%1 BuildMod Save LineNumbers "
                                               "Begin: %2, Action: %3, End: %4, ModPieces: %5")
                                               .arg(edit ? "Update" : "Create")
                                               .arg(SaveModBeginLineNum)
                                               .arg(SaveModActionLineNum)
                                               .arg(SaveModEndLineNum)
                                               .arg(SaveModPieces));
#endif

            // BuildMod meta command lines are written in a bottom up manner

            // Set ModBeginLineNum to the top of the BuildMod step plus the number of 'AddedPieces' introduced by the BuildMod command
            // This is the position for BUILD_MOD END
            ModBeginLineNum += SaveModPieces;

            // Set ModActionLineNum to SaveModBeginLineNum - initial BuildMod insertion line
            // This is the position of BUILD_MOD MOD_END
            ModActionLineNum = SaveModBeginLineNum;

            BuildModData buildModData;

            Where modHere;

            if (BuildModPieces) {

                buildModData = lpub->currentStep->buildMod.value();

                QString metaString;
                buildModData.buildModKey = QString();

                beginMacro("BuildModCreate|" + lpub->viewerStepKey);

                // Delete old BUILD_MOD END meta command
                Where endMod = Where(ModelName, SaveModEndLineNum);
                QString modLine = readLine(endMod);
                Rc rc = lpub->page.meta.parse(modLine, endMod);
                if (rc == BuildModEndRc)
                    deleteLine(endMod);

                // Delete old BUILD_MOD content from bottom up including END_MOD meta command
                endMod = Where(ModelName, SaveModActionLineNum);
                modLine = readLine(endMod);
                rc = lpub->page.meta.parse(modLine, endMod);
                if (rc == BuildModEndModRc)
                    for (modHere = endMod; modHere >= SaveModBeginLineNum; --modHere)
                        deleteLine(modHere);

                // Write BUILD_MOD END meta command at the BuildMod insert position
                modHere = Where(ModelName, ModBeginLineNum);
                buildModData.action      = QString("END");
                lpub->currentStep->buildMod.setValue(buildModData);
                metaString = lpub->currentStep->buildMod.format(false/*local*/,false/*global*/);
                insertLine(modHere, metaString, nullptr);

                // Write BUILD_MOD END_MOD meta command above LPub content first line - last to first
                modHere = Where(ModelName, ModActionLineNum);
                buildModData.action      = QString("END_MOD");
                lpub->currentStep->buildMod.setValue(buildModData);
                metaString = lpub->currentStep->buildMod.format(false,false);
                insertLine(modHere, metaString, nullptr);

                // Write buildMod content last to first
                for (int i = BuildModPieces - 1; i >= 0; --i) {
                    metaString = ViewerModContents.at(i);
                    insertLine(modHere, metaString, nullptr);
                }

                // Write BUILD_MOD BEGIN meta command above buildMod content first line
                buildModData.action      = QString("BEGIN");
                buildModData.buildModKey = BuildModKey;
                lpub->currentStep->buildMod.setValue(buildModData);
                metaString = lpub->currentStep->buildMod.format(false,false);
                insertLine(modHere, metaString, nullptr);

                clearWorkingFiles(getPathsFromViewerStepKey(lpub->viewerStepKey));

                endMacro();
            }

            // BuildMod attribute values are calculated in a top down manner

            // Reset to BUILD_MOD BEGIN
            ModBeginLineNum = SaveModBeginLineNum;

            // Reset to BUILD_MOD END_MOD command line number - add 1 for meta command
            ModActionLineNum = ModBeginLineNum + BuildModPieces + 1;

            // Set to BUILD_MOD END command line number - add 1 for meta command
            ModEndLineNum = ModActionLineNum + SaveModPieces + 1;

            // Set to updated number of pieces in the current Step
            ModStepPieces = mViewerPieces.GetSize();

            QVector<int> ModAttributes =
            {
                ModBeginLineNum,            // 0 BM_BEGIN_LINE_NUM
                ModActionLineNum,           // 1 BM_ACTION_LINE_NUM
                ModEndLineNum,              // 2 BM_END_LINE_NUM
                ModDisplayPageNum,          // 3 BM_DISPLAY_PAGE_NUM
                ModStepPieces,              // 4 BM_STEP_PIECES
                ModelIndex,                 // 5 BM_MODEL_NAME_INDEX
                ModStepLineNum,             // 6 BM_MODEL_LINE_NUM
                ModStepNum                  // 7 BM_MODEL_STEP_NUM
            };

            insertBuildMod(BuildModKey,
                           ModAttributes,
                           ModStepIndex);   // Unique ID

#ifdef QT_DEBUG_MODE
            emit messageSig(LOG_DEBUG, QString("Create BuildMod StepIndex: %1, "
                                               "Action: Apply(64) - %2, "
                                               "Attributes: %3 %4 %5 %6 %7 %8 %9 %10, "
                                               "ModKey: %11, "
                                               "Level: %12")
                                               .arg(edit ? "Update" : "Create")     // 01
                                               .arg(ModStepIndex)                   // 02
                                               .arg(ModBeginLineNum)                // 03 - 0 BM_BEGIN_LINE_NUM
                                               .arg(ModActionLineNum)               // 04 - 1 BM_ACTION_LINE_NUM
                                               .arg(ModEndLineNum)                  // 05 - 2 BM_END_LINE_NUM
                                               .arg(ModDisplayPageNum)              // 06 - 3 BM_DISPLAY_PAGE_NUM
                                               .arg(ModStepPieces)                  // 08 - 4 BM_STEP_PIECES
                                               .arg(ModelIndex)                     // 07 - 5 BM_MODEL_NAME_INDEX
                                               .arg(ModStepLineNum)                 // 08 - 6 BM_MODEL_LINE_NUM
                                               .arg(ModStepNum)                     // 09 - 7 BM_MODEL_STEP_NUM
                                               .arg(ModStepKey)                     // 10
                                               .arg(BuildModKey));                  // 11
#endif

            // Reset the build mod range
            mBuildModRange = { 0/*BM_BEGIN_LINE_NUM*/, 0/*BM_ACTION_LINE_NUM*/, -1/*BM_MODEL_INDEX*/ };
        } // mBuildModRange || edit

        emit progressPermStatusRemoveSig();
        enableBuildModActions();
    }
}

void Gui::applyBuildModification()
{
    if (!lpub->currentStep || exporting())
        return;

    QStringList buildModKeys;
    if (!buildModsCount()) {
        return;
    } else if (buildModsCount() == 1) {
        buildModKeys = getBuildModsList();
    } else {
        BuildModDialogGui *buildModDialogGui = new BuildModDialogGui();
        buildModDialogGui->getBuildMod(buildModKeys, BuildModApplyRc);
    }
    const QString buildModKey = buildModKeys.first();

    emit messageSig(LOG_INFO_STATUS, QString("Processing build modification 'apply' action..."));

    Where topOfStep = lpub->currentStep->topOfStep();

    // get the last action for this build mod
    Rc buildModAction = Rc(getBuildModAction(buildModKey, getBuildModStepIndex(topOfStep)));
    // was the last action defined in this step ?
    Rc buildModStep =  Rc(getBuildModStep(topOfStep));
    // set flag to remove the last action command if it is not 'Apply'
    bool removeActionCommand = buildModStep == buildModAction;

    QString model = topOfStep.modelName;
    QString line = QString::number(topOfStep.lineNumber);
    QString step = QString::number(lpub->currentStep->stepNumber.number);
    QString text, type, title;
    if (getBuildModStepKeyModelIndex(buildModKey) == getSubmodelIndex(model) && getBuildModStepKeyStepNum(buildModKey) > step.toInt()) {
            text  = "Build modification '" + buildModKey + "' was created after this step (" + step + "), "
                    "model '" + model + "', at line " + line + ".<br>"
                    "Applying a build modification before it is created is not supported.<br><br>No action taken.<br>";
            type  = "apply build modification error message";
            title = "Build Modification";

            Preferences::MsgID msgID(Preferences::BuildModErrors, Where("Apply_Before_" + model,line).nameToString());
            Preferences::showMessage(msgID, text, title, type);

            return;

    } else if (getBuildModStepKey(buildModKey) == lpub->currentStep->viewerStepKey) {
        text  = "Build modification '" + buildModKey + "' was created in this step (" + step + "), "
                "model '" + model + "', at line " + line + ".<br>"
                "It was automatically applied to the step it was created in.<br><br>No action taken.<br>";
        type  = "apply build modification error message";
        title = "Build Modification";

        Preferences::MsgID msgID(Preferences::BuildModErrors, Where("Source_" + model,line).nameToString());
        Preferences::showMessage(msgID, text, title, type);

        return;

    } else if (buildModAction == BuildModApplyRc) {
        text  = "Build modification '" + buildModKey + "' was already applied to step (" + step + "), "
                "model '" + model + ".<br><br>No action taken.<br>";
        type  = "apply build modification error message";
        title = "Build Modification";

        Preferences::MsgID msgID(Preferences::BuildModErrors, Where("Already_Applied_" + model,line).nameToString());
        Preferences::showMessage(msgID, text, title, type);

        return;

    }  /* else {
        text  = "This action will apply build modification '" + buildModKey + "' "
                "beginning at step " + step + ", in model '" + model + "'.<br><br>Are you sure ?<br>";
        type  = "apply build modification message";
        title = "Build Modification";

        Preferences::MsgID msgID(Preferences::BuildModErrors, Where(model,line).nameToString());
        switch (Preferences::showMessage(msgID, text, title, type, true / *option* /))
        {
        default:
        case QMessageBox::Cancel:
            return; // cancel request
        case QMessageBox::Ok:
            break;  // apply changes
        }


    } */

    int it = lcGetActiveProject()->GetImageType();
    if (it == Options::CSI) {
        QString metaString;
        bool newCommand = false;
        BuildModData buildModData = lpub->currentStep->buildMod.value();

        beginMacro("BuildModApply|" + lpub->currentStep->viewerStepKey);

        buildModData.action      = QString("APPLY");
        buildModData.buildModKey = buildModKey;
        lpub->currentStep->buildMod.setValue(buildModData);
        metaString = lpub->currentStep->buildMod.format(false/*local*/,false/*global*/);
        newCommand = lpub->currentStep->buildMod.here() ==  Where();
        lpub->currentStep->mi(it)->setMetaAlt(newCommand ? topOfStep : lpub->currentStep->buildMod.here(), metaString, newCommand, removeActionCommand);

        if (removeActionCommand)
            clearBuildModAction(buildModKey, getBuildModStepIndex(topOfStep));

        clearWorkingFiles(getPathsFromViewerStepKey(lpub->currentStep->viewerStepKey));

        endMacro();
    }
}

void Gui::removeBuildModification()
{
    if (!lpub->currentStep || exporting())
        return;

    QStringList buildModKeys;
    if (!buildModsCount()) {
        return;
    } else if (buildModsCount() == 1) {
        buildModKeys = getBuildModsList();
    } else {
        BuildModDialogGui *buildModDialogGui =
                new BuildModDialogGui();
        buildModDialogGui->getBuildMod(buildModKeys, BuildModRemoveRc);
    }
    const QString buildModKey = buildModKeys.first();

    emit messageSig(LOG_INFO_STATUS, QString("Processing build modification 'remove' action..."));

    Where topOfStep = lpub->currentStep->topOfStep();

    // get the last action for this build mod
    Rc buildModAction = Rc(getBuildModAction(buildModKey, getBuildModStepIndex(topOfStep)));
    // was the last action defined in this step ?
    Rc buildModStep =  Rc(getBuildModStep(topOfStep));
    // set flag to remove the last action command if it is not 'Remove'
    bool removeActionCommand = buildModStep == buildModAction;

    QString model = topOfStep.modelName;
    QString line = QString::number(topOfStep.lineNumber);
    QString step = QString::number(lpub->currentStep->stepNumber.number);
    QString text, type, title;
    if (getBuildModStepKeyModelIndex(buildModKey) == getSubmodelIndex(model) && getBuildModStepKeyStepNum(buildModKey) > step.toInt()) {
            text  = "Build modification '" + buildModKey + "' was created after this step (" + step + "), "
                    "model '" + model + "', at line " + line + ".<br>"
                    "Removing a build modification before it is created is not supported.<br><br>No action taken.<br>";
            type  = "remove build modification error message";
            title = "Build Modification";

            return;

            Preferences::MsgID msgID(Preferences::BuildModErrors, Where("Remove_Before_" + model,line).nameToString());
            Preferences::showMessage(msgID, text, title, type);
    } else if (getBuildModStepKey(buildModKey) == lpub->viewerStepKey) {
        text  = "Build modification '" + buildModKey + "' was created in this step (" + step + "), "
                "in model '" + model + "' at line " + line + ".<br><br>"
                "It cannot be removed from the step it was created in.<br><br>"
                "Select 'Delete Build Modification' to delete from '" + model + "', "
                "step " + step + " at line " + line;
        type  = "remove build modification error message";
        title = "Build Modification";

        Preferences::MsgID msgID(Preferences::BuildModErrors, Where("Not_Applied_" + model,line).nameToString());
        Preferences::showMessage(msgID, text, title, type);

        return;

    } else if (buildModAction == BuildModRemoveRc) {
        text  = "Build modification '" + buildModKey + "' was already removed from step (" + step + "), "
                "model '" + model + ".<br><br>No action taken.<br>";
        type  = "remove build modification error message";
        title = "Build Modification";

        Preferences::MsgID msgID(Preferences::BuildModErrors, Where("Already_Removed_" + model,line).nameToString());
        Preferences::showMessage(msgID, text, title, type);

        return;

    } /* else {
        text  = "This action will remove build modification '" + buildModKey + "' "
                "beginning at step " + step + " in model '" + model + "'.<br><br>Are you sure ?<br>";
        type  = "remove build modification message";
        title = "Build Modification";

        Preferences::MsgID msgID(Preferences::BuildModErrors, Where(model,line).nameToString());
        switch (Preferences::showMessage(msgID, text, title, type, true / *option* /))
        {
        default:
        case QMessageBox::Cancel:
            return;
        case QMessageBox::Ok:
            break;
        }
    } */

    int it = lcGetActiveProject()->GetImageType();
    if (it == Options::CSI) {
        QString metaString;
        bool newCommand = false;
        BuildModData buildModData = lpub->currentStep->buildMod.value();

        beginMacro("BuildModRemove|" + lpub->currentStep->viewerStepKey);

        buildModData.action      = QString("REMOVE");
        buildModData.buildModKey = buildModKey;
        lpub->currentStep->buildMod.setValue(buildModData);
        metaString = lpub->currentStep->buildMod.format(false/*local*/,false/*global*/);
        newCommand = lpub->currentStep->buildMod.here() == Where();
        lpub->currentStep->mi(it)->setMetaAlt(newCommand ? topOfStep : lpub->currentStep->buildMod.here(), metaString, newCommand, removeActionCommand);

        if (removeActionCommand)
            clearBuildModAction(buildModKey, getBuildModStepIndex(topOfStep));

        clearWorkingFiles(getPathsFromViewerStepKey(lpub->currentStep->viewerStepKey));

        endMacro();
    }

    enableBuildModMenuAndActions();
}

void Gui::loadBuildModification()
{
    if (exporting())
        return;

    QStringList buildModKeys;
    if (!buildModsCount()) {
        return;
    } else if (buildModsCount() == 1) {
        buildModKeys = getBuildModsList();
    } else {
        BuildModDialogGui *buildModDialogGui =
                new BuildModDialogGui();
        buildModDialogGui->getBuildMod(buildModKeys, BM_CHANGE);
    }
    const QString buildModKey = buildModKeys.first();

    emit messageSig(LOG_INFO_STATUS, QString("Processing build modification 'load' action..."));

    /*
    QString model = "undefined", line = "undefined", step = "undefined";
    QStringList keys = getViewerStepKeys(true/ *get Name* /, false/ *pliPart* /, getBuildModStepKey(buildModKey));
    if (keys.size() > 2) { model = keys[0]; line = keys[1]; step = keys[2]; }
    QString text  = "This action will load build modification '" + buildModKey + "' "
                     ", step " + step + ", model '" + model + "' into the Visual Editor "
                    "to allow editing.<br><br>Are you sure ?<br>";
    QString type  = "load build modification";
    QString title = "Build Modification";
    Preferences::MsgID msgID(Preferences::BuildModErrors, Where(model,line).nameToString());
    switch (Preferences::showMessage(msgID, text, title, type, true / *option* /))
    {
    default:
    case QMessageBox::Cancel:
        return;
    case QMessageBox::Ok:
        break;
    }
    */

    int it = lcGetActiveProject()->GetImageType();
    if (it == Options::CSI) {

        buildModChangeKey = "";

        int buildModDisplayPageNum = getBuildModDisplayPageNumber(buildModKey);

        QString buildModStepKey = getBuildModStepKey(buildModKey);

        if (buildModDisplayPageNum && ! buildModStepKey.isEmpty()) {

            if (buildModDisplayPageNum != displayPageNum) {
                displayPageNum = buildModDisplayPageNum;
                displayPage();
            }

            bool setBuildModStep = lpub->currentStep && lpub->currentStep->viewerStepKey != buildModStepKey;

            if (isViewerStepMultiStep(buildModStepKey) && setBuildModStep) {
                if (lpub->setCurrentStep(buildModStepKey)) {
                    showLine(lpub->currentStep->topOfStep());
                    lpub->currentStep->loadTheViewer();
                }
            }

            buildModChangeKey = buildModKey;
        }
    }
}

bool Gui::setBuildModChangeKey()
{
    if (!lpub->currentStep)
        return false;

    int it = lcGetActiveProject()->GetImageType();
    if (it == Options::CSI) {
        Rc rc;
        Where walk = lpub->currentStep->top;

        QString line = readLine(walk);
        rc =  lpub->page.meta.parse(line,walk,false);
        if (rc == StepRc || rc == RotStepRc)
            walk++;   // Advance past STEP meta

        // Parse the step lines
        for ( ;
              walk.lineNumber < subFileSize(walk.modelName);
              walk.lineNumber++) {
            line = readLine(walk);
            Where here(walk.modelName,walk.lineNumber);
            rc =  lpub->page.meta.parse(line,here,false);

            switch (rc) {
            case BuildModBeginRc:
                buildModChangeKey = lpub->page.meta.LPub.buildMod.key();
                return true;

            // Search until next step/rotstep meta
            case RotStepRc:
            case StepRc:
                return false;

            default:
                break;
            }
        }
    }

    return false;
}

void Gui::updateBuildModification()
{
    if (buildModChangeKey.isEmpty()) {
        setBuildModChangeKey();
        createBuildModification();
    } else {
        createBuildModification();
    }
}

void Gui::deleteBuildModification()
{
    QStringList buildModKeys;
    if (!buildModsCount()) {
        return;
    } else if (buildModsCount() == 1) {
        buildModKeys = getBuildModsList();
    } else {
        BuildModDialogGui *buildModDialogGui =
                new BuildModDialogGui();
        buildModDialogGui->getBuildMod(buildModKeys, BM_DELETE);
    }
    const QString buildModKey = buildModKeys.first();

    emit messageSig(LOG_INFO_STATUS, QString("Processing build modification 'delete' action..."));

    QString model = "undefined", line = "undefined", step = "undefined";
    QStringList keys = LPub::getViewerStepKeys(true/*get Name*/, false/*pliPart*/, lpub->ldrawFile.getBuildModStepKey(buildModKey));
    if (keys.size() > 2) { model = keys[0]; line = keys[1]; step = keys[2]; }
    /*
    QString text  = "This action will permanently delete build modification '" + buildModKey + "' "
                    "from '" + model + "' at step " + step + "' and cannot be completelly undone using the Undo menu action.<br><br>"
                    "The modified CSI image and Visual Editor entry will be parmanently deleted.<br>"
                    "However, you can use 'Reload' menu action to restore all deleted content.<br><br>"
                    "Do you want to continue ?<br>";
    QString type  = "delete build modification";
    QString title = "Build Modification";

    Preferences::MsgID msgID(Preferences::BuildModErrors, Where(model,line).nameToString());
    switch (Preferences::showMessage(msgID, text, title, type, true / *option* /))
    {
    default:
    case QMessageBox::Cancel:
        return;
    case QMessageBox::Ok:
        break;
    }
    */

    const QString buildModStepKey = getBuildModStepKey(buildModKey);
    bool multiStepPage = isViewerStepMultiStep(buildModStepKey);

    // Delete options
    QPixmap _icon = QPixmap(":/icons/lpub96.png");
    QMessageBoxResizable box;
    box.setWindowIcon(QIcon());
    box.setIconPixmap (_icon);
    box.setWindowFlags (Qt::Dialog | Qt::CustomizeWindowHint | Qt::WindowTitleHint);

    QString title = tr("Build Modification Image");
    QString text = tr("Click 'Submodel',%1 or 'Step' to reset the respective image cache.").arg(multiStepPage? " 'Page',": "");

    box.setWindowTitle(QString("%1 %2").arg(VER_PRODUCTNAME_STR).arg(title));
    box.setText (tr("Select your option to reset the image cache."));
    box.setInformativeText(text);

    QPushButton *clearSubmodelButton = box.addButton(tr("Modified"), QMessageBox::AcceptRole);
    clearSubmodelButton->setToolTip(tr("Reset modified submodel images from step %1").arg(step));

    QPushButton *clearPageButton = nullptr;
    if (multiStepPage) {
        int pageNum = getBuildModDisplayPageNumber(buildModKey);
        QPushButton *clearPageButton = box.addButton(tr("Page %1").arg(pageNum), QMessageBox::AcceptRole);
        clearPageButton->setToolTip(tr("Reset page %1 step group images").arg(pageNum));
    }

    QPushButton *clearStepButton = box.addButton(tr("Step %1").arg(step), QMessageBox::AcceptRole);
    clearStepButton->setToolTip(tr("Reset step %1 image only").arg(step));

    QPushButton *cancelButton = box.addButton(QMessageBox::Cancel);
    cancelButton->setDefault(true);

    box.exec();

    bool clearSubmodel = box.clickedButton() == clearSubmodelButton;
    bool clearPage     = multiStepPage ? box.clickedButton() == clearPageButton : false;
    bool clearStep     = box.clickedButton() == clearStepButton;

    if (box.clickedButton() == cancelButton)
        return;

    QString clearOption = clearSubmodel ? "_cm" : clearPage ? "_cp" : clearStep ? "_cs" : QString();

    int it = lcGetActiveProject()->GetImageType();
    if (it == Options::CSI) {
        int modBeginLineNum  = getBuildModBeginLineNumber(buildModKey);
        int modActionLineNum = getBuildModActionLineNumber(buildModKey);
        int modEndLineNum    = getBuildModEndLineNumber(buildModKey);
        QString modelName    = getBuildModStepKeyModelName(buildModKey);

        if (modelName.isEmpty() || !modBeginLineNum || !modActionLineNum || !modEndLineNum) {
            emit messageSig(LOG_ERROR, QString("There was a problem receiving build mod attributes for key [%1]<br>"
                                               "Delete build modification cannot continue.").arg(buildModKey));
            return;
        }

        beginMacro("BuildModDelete|" + buildModStepKey + clearOption);

        // delete existing APPLY/REMOVE (action) commands, starting from the bottom of the step
        Rc rc;
        QString modKey, modLine;
        Where here, topOfStep, bottomOfStep;
        QMap<int, int> actionsMap = getBuildModActions(buildModKey);
        QList<int> stepIndexes = actionsMap.keys();
        std::sort(stepIndexes.begin(), stepIndexes.end(), std::greater<int>()); // sort stepIndexes descending
        for (int &stepIndex : stepIndexes) {
            QMap<int, int>::iterator i = actionsMap.find(stepIndex);
            if (i.key()) {                                                      // skip first step at index 0 - deleted later
                if (getBuildModStepIndexWhere(i.key(), topOfStep)) {            // initialize topOfStep Where
                    if (! getBuildModStepIndexWhere(i.key() + 1, bottomOfStep)) // bottom of step is top of next step
                        // handle last step
                        bottomOfStep = Where(topOfStep.modelName, subFileSize(topOfStep.modelName));
                    for (Where walk = bottomOfStep; walk > topOfStep.lineNumber; --walk) {
                        here = walk;
                        modLine = readLine(here);
                        rc = lpub->page.meta.parse(modLine, here);
                        switch (rc) {
                        case BuildModApplyRc:
                        case BuildModRemoveRc:
                            modKey = lpub->page.meta.LPub.buildMod.key();
                            if (modKey == buildModKey)
                                deleteLine(here);
                            break;
                        default:
                            break;
                        }
                    }
                }
            }
        }

        // delete existing BUILD_MOD commands from bottom up, starting at END
        here = Where(modelName, modEndLineNum);
        modLine = readLine(here);
        rc = lpub->page.meta.parse(modLine, here);
        if (rc == BuildModEndRc)
            deleteLine(here);

        // delete existing BUILD_MOD commands from bottom up, starting at END_MOD
        here = Where(modelName, modActionLineNum);
        modLine = readLine(here);
        rc = lpub->page.meta.parse(modLine, here);
        if (rc == BuildModEndModRc)
            for (Where walk = here; walk >= modBeginLineNum; --walk)
                deleteLine(walk);

        // delete step image to trigger image regen
        if (clearSubmodel) {
            clearWorkingFiles(getPathsFromViewerStepKey(buildModStepKey));
        } else if (clearPage) {
            PlacementType relativeType = multiStepPage ? StepGroupType : SingleStepType;
            clearPageCache(relativeType, &lpub->page, Options::CSI);
        } else if (clearStep) {
            QString csiPngName = getViewerStepImagePath(buildModStepKey);
            clearStepCSICache(csiPngName);
            // delete viewer step to trigger viewer update
            if (!deleteViewerStep(buildModStepKey))
                emit messageSig(LOG_ERROR,QString("Failed to delete viewer step entry for key %1.").arg(buildModStepKey));
        }

        // delete BuildMod
        if (!deleteBuildMod(buildModKey))
            emit messageSig(LOG_ERROR,QString("Failed to delete build modification for key %1.").arg(buildModKey));

        endMacro();
    }

    enableBuildModMenuAndActions();
}

/*********************************************
 *
 * save viewer model if modified
 *
 ********************************************/

bool Gui::saveBuildModification()
{
    if (!Preferences::buildModEnabled)
        return true;     // continue

    Project* Project = lcGetActiveProject();
    if (Project->GetImageType() != Options::CSI)
        return true;     // continue

    QString change = undoAct->data().toString();

    if (!BuildModChangeTriggers.contains(change))
        return true;     // continue

    QPixmap _icon = QPixmap(":/icons/lpub96.png");
    QMessageBox box;
    box.setWindowIcon(QIcon());
    box.setIconPixmap (_icon);
    box.setTextFormat (Qt::RichText);
    box.setWindowTitle(tr ("%1 Save Model Change").arg(VER_PRODUCTNAME_STR));
    box.setWindowFlags (Qt::Dialog | Qt::CustomizeWindowHint | Qt::WindowTitleHint);
    QString title = "<b>" +
            (setBuildModChangeKey()
             ? tr("Save model changes to build modification '%1'?").arg(getBuildModChangeKey())
             : tr("Save model changes as build modification?")) + "</b>";
    QString text = tr("Save changes as a build modification to this step?");
    box.setText (title);
    box.setInformativeText (text);
    box.setStandardButtons (QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
    box.setDefaultButton   (QMessageBox::Cancel);
    switch (box.exec())
    {
    default:
    case QMessageBox::Cancel:
        return false; // cancel request

    case QMessageBox::Yes:
        updateBuildModification();
        break;       // apply changes

    case QMessageBox::No:
        break;       // discard changes and continue
    }

    return true;
}

/*********************************************
 *
 * set viewer step key
 *
 ********************************************/

void Gui::setViewerStepKey(const QString &stepKey, int imageType)
{
    Q_UNUSED(imageType)

    lpub->viewerStepKey = stepKey;

    mBuildModRange = { 0, 0, -1 };
}

/*********************************************
 *
 * set step from specified line
 *
 ********************************************/

void Gui::setStepForLine()
{
    if (!lpub->currentStep || !gMainWindow || !gMainWindow->isVisible() || exporting())
        return;

    disconnect(this,           SIGNAL(highlightSelectedLinesSig(QVector<int> &, bool)),
               editWindow,     SLOT(  highlightSelectedLines(   QVector<int> &, bool)));

    enableBuildModActions();
    getCurrentStep()->viewerOptions->ZoomExtents = true;
    getCurrentStep()->loadTheViewer();

    connect(this,           SIGNAL(highlightSelectedLinesSig(QVector<int> &, bool)),
            editWindow,     SLOT(  highlightSelectedLines(   QVector<int> &, bool)));
}

/*********************************************
 *
 * slelcted Line
 *
 ********************************************/

bool Gui::getSelectedLine(int modelIndex, int lineIndex, int source, int &lineNumber) {

    lineNumber        = EDITOR_LINE   ;             // 0
    bool currentModel = modelIndex == QString(lpub->viewerStepKey[0]).toInt();
    bool newLine      = lineIndex  == NEW_PART;     //-1
    bool fromViewer   = source      > EDITOR_LINE;  // 0

//#ifdef QT_DEBUG_MODE
//    emit messageSig(LOG_DEBUG, QString("currentModel: %1 [modelIndex: %2 == ViewerStepKey(index,line,step)[0]: %3]")
//                    .arg(currentModel ? "True" : "False")
//                    .arg(modelIndex)
//                    .arg(lpub->viewerStepKey));
//    emit messageSig(LOG_DEBUG, QString("newLine: %1 [lineIndex: %2 == NEW_PART: -1]")
//                    .arg(newLine ? "True" : "False")
//                    .arg(lineIndex));
//    emit messageSig(LOG_DEBUG, QString("fromViewer: %1 [source: %2 > EDITOR_LINE: 0]")
//                    .arg(fromViewer ? "True" : "False")
//                    .arg(source));
//#endif

    if (newLine) {
        emit messageSig(LOG_TRACE, QString("New viewer part modelName [%1]")
                                           .arg(getSubmodelName(modelIndex)));
        return false;

    } else if (currentModel) {

        if (!lpub->currentStep)
            return false;

#ifdef QT_DEBUG_MODE
        emit messageSig(LOG_TRACE, QString("LPub Step lineIndex count: %1 item(s)")
                        .arg(lpub->currentStep->lineTypeIndexes.size()));
//      for (int i = 0; i < lpub->currentStep->lineTypeIndexes.size(); ++i)
//          emit messageSig(LOG_TRACE, QString(" -LPub Part lineNumber [%1] at step line lineIndex [%2] - specified lineIndex [%3]")
//                                             .arg(lpub->currentStep->lineTypeIndexes.at(i)).arg(i).arg(lineIndex));
#endif

        if (fromViewer)      // input relativeIndes
            lineNumber = lpub->currentStep->getLineTypeRelativeIndex(lineIndex);
        else                 // input lineTypeIndex
            lineNumber = lpub->currentStep->getLineTypeIndex(lineIndex);

    } else if (modelIndex != NEW_MODEL) {

        if (fromViewer)      // input relativeIndes
            lineNumber = getLineTypeRelativeIndex(modelIndex,lineIndex); // return lineTypeIndex - part lineNumber
        else                 // input lineTypeIndex
            lineNumber = getLineTypeIndex(modelIndex,lineIndex);         // return relativeIndex - step line lineIndex
    } else
        return false;        // lineNumber = EDITOR_LINE;

//#ifdef QT_DEBUG_MODE
//    emit messageSig(LOG_DEBUG, QString("selected lineNumber: [%1] > NOT_FOUND: [-1]")
//                    .arg(lineNumber));
//#endif

    return lineNumber > NOT_FOUND; // -1
}

/*********************************************
 *
 * Selected part index(es)
 *
 ********************************************/

void Gui::SelectedPartLines(QVector<TypeLine> &indexes, PartSource source) {
    if (! exporting()) {
        if (!lpub->currentStep || (source == EDITOR_LINE && !indexes.size()))
            return;

        QVector<int> lines;
        bool fromViewer   = source > EDITOR_LINE;
        bool validLine    = false;
        bool modsEnabled  = false;
        int lineNumber    = 0;
        int lineIndex     = NEW_PART;
        int modelIndex    = NEW_MODEL;
        QString modelName = "undefined";

#ifdef QT_DEBUG_MODE
        const QString SourceNames[] =
        {
            "EDITOR_CLR",  // -2
            "NOT_FOUND",   // -1 OUT_OF_BOUNDS, NEW_PART, NEW_MODEL
            "VIEWER_NONE", //  0 EDITOR_LINE
            "VIEWER_LINE", //  1
            "VIEWER_MOD",  //  2
            "VIEWER_DEL",  //  3
            "VIEWER_SEL",  //  4
            "VIEWER_CLR"   //  5
        };
        QString fromSource;
        if (!fromViewer && source == EDITOR_LINE)
            fromSource = "EDITOR_LINE";
        else
            fromSource = SourceNames[source+2];
        emit gui->messageSig(LOG_DEBUG, QString("Selection Source: %1 (%2)")
                                                .arg(fromSource, QString::number(source)));
#endif
        if (indexes.size()) {
            modelName  = getSubmodelName(indexes.at(0).modelIndex);
            modelIndex = indexes.at(0).modelIndex;
        } else if (!lpub->viewerStepKey.isEmpty()) {
            modelName  = lpub->currentStep->topOfStep().modelName;
            modelIndex = getSubmodelIndex(modelName);
        }

#ifdef QT_DEBUG_MODE
        if (modelIndex != NEW_MODEL && source > VIEWER_LINE)
            emit messageSig(LOG_TRACE, QString("Submodel lineIndex count: %1 item(s)")
                            .arg(lpub->ldrawFile.getLineTypeRelativeIndexCount(modelIndex)));
#endif

        for (int i = 0; i < indexes.size() && source < VIEWER_CLR && source > EDITOR_CLR; ++i) {
            lineIndex = indexes.at(i).lineIndex;
            // New part lines are added in createBuildModification() routine
            if (lineIndex != NEW_PART) {
                validLine = getSelectedLine(modelIndex, lineIndex, source, lineNumber);
                if (lineNumber != OUT_OF_BOUNDS)
                    lines.append(lineNumber);
            }

            if (validLine) {
                bool buildModChange = source == VIEWER_MOD || source == VIEWER_DEL;
                if (fromViewer && buildModChange && Preferences::buildModEnabled) {
                    if (mBuildModRange.first()) {
                        if (lineNumber < mBuildModRange.first())
                            mBuildModRange[BM_BEGIN_LINE_NUM] = lineNumber;
                        else if (lineNumber > mBuildModRange.last())
                            mBuildModRange[BM_ACTION_LINE_NUM] = lineNumber;
                        mBuildModRange[BM_MODEL_INDEX] = modelIndex;
                    } else {
                        mBuildModRange = { lineNumber, lineNumber, modelIndex };
                    }
                    if (!modsEnabled) {
                        CreateBuildModAct->setEnabled(true);
                        enableBuildModification();
                        modsEnabled = true;
                    }
                }
            }

#ifdef QT_DEBUG_MODE
            QString Message;
            if (fromViewer) {
                if (lineIndex == NEW_PART) {
                    Message = tr("New viewer part specified at step %1, modelName: [%2]")
                            .arg(lpub->currentStep->stepNumber.number)
                            .arg(modelName);
                } else if (validLine) {
                    Message = tr("Selected part modelName [%1] lineNumber: [%2] at step line index [%3]")
                            .arg(modelName).arg(lineNumber).arg(lineIndex < 0 ? "undefined" : QString::number(lineIndex));
                } else {
                    Message = tr("%1 part lineNumber [%2] for step line index [%3]")
                            .arg(indexes.size() ? "Out of bounds" : "Invalid")
                            .arg(lineNumber).arg(lineIndex < 0 ? "undefined" : QString::number(lineNumber));
                }
            } else if (validLine) { // valid and not from viewer
                Message = tr("Selected part modelName [%1] lineNumber: [%2] at step line index [%3]")
                        .arg(modelName).arg(lineIndex).arg(lineNumber < 0 ? "undefined" : QString::number(lineIndex));
            } else {                // invalid and not from viewer
                Message = tr("%1 part lineNumber [%2] for step line index [%3]") // index and number flipped
                        .arg(indexes.size() ? "Out of bounds" : "Invalid")
                        .arg(lineIndex).arg(lineNumber < 0 ? "undefined" : QString::number(lineNumber));
            }
            emit messageSig(LOG_TRACE, Message);
#endif
        } // indexes present and source is not VIEWER_CLR

        if (fromViewer) {
            if (editWindow->isVisible() && (source == VIEWER_SEL || source == VIEWER_MOD || source == VIEWER_CLR)) {
                emit highlightSelectedLinesSig(lines, source == VIEWER_CLR);
            }
            // delete action with no selected lines
            if (source == VIEWER_DEL && Preferences::buildModEnabled) {
                if (!modsEnabled) {
                    CreateBuildModAct->setEnabled(true);
                    enableBuildModification();
                }
                emit messageSig(LOG_TRACE, tr("Delete viewer part(s) specified at step %1, modelName: [%2]")
                                              .arg(lpub->currentStep->stepNumber.number)
                                              .arg(modelName));
            }
        } else { // indexes from editor
            if (lines.size() && gMainWindow->isVisible())
                emit setSelectedPiecesSig(lines);
        }
    } // not exporting
}

void Gui::importLDD()
{
    if (!maybeSave() || !saveBuildModification())
        return;

    QString LoadFileName = QFileDialog::getOpenFileName(this, tr("Import"), QString(), tr("LEGO Diginal Designer Files (*.lxf);;All Files (*.*)"));
    if (LoadFileName.isEmpty())
        return;

    Project* Importer = new Project();

    if (Importer->ImportLDD(LoadFileName))
    {
        QString FileName = QFileDialog::getSaveFileName(this,tr("Save Diginal Designer File As"),
                                                        tr("%1.%2.mpd").arg(QFileInfo(LoadFileName).completeBaseName()),
                                                        tr("LDraw Files (*.mpd *.ldr *.dat);;All Files (*.*)"));
        if (FileName.isEmpty()) {
            QTemporaryDir tempDir;
            if (tempDir.isValid())
                FileName =  QString("%1/%2.%3.mpd").arg(tempDir.path(),QFileInfo(LoadFileName).completeBaseName());
        }
        saveImport(FileName, Importer);
    }
    else
        delete Importer;
}

void Gui::importInventory()
{
    if (!maybeSave() || !saveBuildModification())
        return;

    lcSetsDatabaseDialog Dialog(this);
    if (Dialog.exec() != QDialog::Accepted)
        return;

    Project* Importer = new Project();

    if (Importer->ImportInventory(Dialog.GetSetInventory(), Dialog.GetSetName(), Dialog.GetSetDescription()))
    {
        QString FileName = QFileDialog::getSaveFileName(this,tr("Save Inventory File As"),
                                                        tr("%1.%2.mpd").arg(Dialog.GetSetName()),
                                                        tr("LDraw Files (*.mpd *.ldr *.dat);;All Files (*.*)"));
        if (FileName.isEmpty()) {
            QTemporaryDir tempDir;
            if (tempDir.isValid())
                FileName =  QString("%1/%2.%3.mpd").arg(tempDir.path(),Dialog.GetSetName());
        }
        saveImport(FileName, Importer);
    }
    else
        delete Importer;
}

bool Gui::saveImport(const QString& FileName, Project *Importer)
{
    QFile File(FileName);

    if (!File.open(QIODevice::WriteOnly))
    {
        emit messageSig(LOG_ERROR, tr("Error writing to file '%1':\n%2").arg(FileName, File.errorString()));
        return false;
    }

    QTextStream Stream(&File);
    bool firstStep = true;
    QLatin1String LineEnding("\r\n");
    QLatin1String Extension(".ldr");
    for (lcModel* Model : Importer->GetModels())
    {
        if (firstStep) {
            firstStep = false;
            Stream << QLatin1String("0 FILE Import-") << Model->GetProperties().mFileName << Extension << LineEnding;
            Stream << QLatin1String("0 ") << Model->GetProperties().mDescription << LineEnding;
            Stream << QLatin1String("0 Name: Import-") << Model->GetProperties().mModelName << LineEnding;
            Stream << QLatin1String("0 Author: ") << Model->GetProperties().mAuthor << LineEnding;
            Stream << QLatin1String("0 !LPUB INSERT TEXT \"This page is intentionally left blank\\nto avoid loading the imported parts at the first step.\" \"Arial,36,-1,255,75,0,0,0,0,0\" \"Black\" OFFSET 0.0219 0.4814\r\n");
            Stream << QLatin1String("0 !LPUB INSERT PAGE\r\n");
            Stream << QLatin1String("0 STEP\r\n");
            Stream << QLatin1String("1 16 0 0 0 1 0 0 0 1 0 0 0 1 ") << Model->GetProperties().mFileName << Extension << LineEnding;
            Stream << QLatin1String("0 NOFILE\r\n");
        }
        Stream << QLatin1String("0 FILE ") << Model->GetProperties().mFileName << Extension << LineEnding;
        Model->SaveLDraw(Stream, false);
        Model->SetSaved();
        Stream << QLatin1String("0 NOFILE\r\n");
    }
    File.close();

    int saveLoadMessageFlag = Preferences::ldrawFilesLoadMsgs;
    Preferences::ldrawFilesLoadMsgs = SHOW_WARNING;
    if (!loadFile(FileName)) {
        Preferences::ldrawFilesLoadMsgs = saveLoadMessageFlag;
        return false;
    }
    Preferences::ldrawFilesLoadMsgs = saveLoadMessageFlag;

    return true;
}

/*********************************************
 *
 * Default camera settings
 *
 *********************************************/

float MetaDefaults::getCameraFOV()
{
    return (Preferences::preferredRenderer == RENDERER_NATIVE ?
            gApplication ? lcGetPreferences().mCFoV : CAMERA_FOV_NATIVE_DEFAULT :
            Preferences::preferredRenderer == RENDERER_LDVIEW && Preferences::perspectiveProjection ?
            CAMERA_FOV_LDVIEW_P_DEFAULT :
            CAMERA_FOV_DEFAULT);
}

float MetaDefaults::getFOVMinRange()
{
    return (Preferences::preferredRenderer == RENDERER_NATIVE ?
            CAMERA_FOV_NATIVE_MIN_DEFAULT :
            CAMERA_FOV_MIN_DEFAULT);
}

float MetaDefaults::getFOVMaxRange()
{
    return (Preferences::preferredRenderer == RENDERER_NATIVE ?
            CAMERA_FOV_NATIVE_MAX_DEFAULT :
            Preferences::preferredRenderer == RENDERER_LDVIEW && Preferences::perspectiveProjection ?
            CAMERA_FOV_LDVIEW_P_MAX_DEFAULT :
            CAMERA_FOV_MAX_DEFAULT);
}

float MetaDefaults::getNativeCameraZNear()
{
    return gApplication ? lcGetPreferences().mCNear : CAMERA_ZNEAR_NATIVE_DEFAULT;
}

float MetaDefaults::getNativeCameraZFar()
{
    return gApplication ? lcGetPreferences().mCFar : CAMERA_ZFAR_NATIVE_DEFAULT;
}

float MetaDefaults::getCameraDDF()
{
    return gApplication ? lcGetPreferences().mDDF : CAMERA_DISTANCE_FACTOR_NATIVE_DEFAULT;
}

QString MetaDefaults::getPreferredRenderer()
{
    const QString rendererNames[NUM_RENDERERS] =
    {
        "Native",  // RENDERER_NATIVE
        "LDView",  // RENDERER_LDVIEW
        "LDGLite", // RENDERER_LDGLITE
        "POVRay"   // RENDERER_POVRAY
    };
    return rendererNames[Preferences::preferredRenderer];
}

QString MetaDefaults::getPreferredUnits()
{
    return Preferences::preferCentimeters ? QObject::tr("dots per centimetre (DPCM) unit") :
                                            QObject::tr("dots per inch (DPI) unit");
}

float MetaDefaults::getAssemblyCameraLatitude()
{
    return (float)Preferences::assemblyCameraLatitude;
}

float MetaDefaults::getAssemblyCameraLongitude()
{
    return (float)Preferences::assemblyCameraLongitude;
}

float MetaDefaults::getPartCameraLatitude()
{
    return (float)Preferences::partCameraLatitude;
}

float MetaDefaults::getPartCameraLongitude()
{
    return (float)Preferences::partCameraLongitude;
}

float MetaDefaults::getSubmodelCameraLatitude()
{
    return (float)Preferences::submodelCameraLatitude;
}

float MetaDefaults::getSubmodelCameraLongitude()
{
    return (float)Preferences::submodelCameraLongitude;
}
